/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.simulation.SimulationResultData;
import de.bsvrz.ars.ars.persistence.CacheManager;
import de.bsvrz.ars.ars.persistence.ContainerFileDirIterator;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.dataSerializer.NoSuchVersionException;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;

import java.io.File;
import java.io.IOException;
import java.util.concurrent.BlockingQueue;

/**
 * Task zum Loeschen von Simulationsdaten einer bestimmten Simulationsvariante. Der Task wird verwendet um von Clients und der Simulation {@link
 * de.bsvrz.ars.ars.mgmt.simulation.SimulationAutomaton} eingehende Löschaufträge zu verarbeiten. Die Daten werden sowohl von dem Filesystem entfernt und die
 * betroffenen DataIdentNodes werden aus dem DIN-Tree genommen.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DeleteSimVarTask extends SingleClientSenderTask {

	/** @see SingleClientSenderTask#SingleClientSenderTask(ArchiveManager,RingOfDeathListener,RingBuffer) */
	public DeleteSimVarTask(ArchiveManager archiveMgr,BlockingQueue<Object> inputData) {
		super(archiveMgr, inputData);
	}

	/** @see de.bsvrz.ars.ars.mgmt.tasks.SingleTask#work(ResultData) */
	protected void work(ResultData resultData) {
		if(resultData != null) {
			try {
				if(resultData instanceof SimulationResultData)
				// Löschauftrag wurde von einer Simulation gestellt
				{
					deleteSimVarSimulation((SimulationResultData)resultData);
				}
				else if(resultData.hasData())
				// Löschauftrag wurde vom Client beauftragt
				{
					deleteSimVarClient(resultData);
				}
			}
			catch(IOException e) {
				String msg = "Fehler beim Deserialisieren. " + e.getMessage();
				_debug.warning(msg);
				sendErrorResponse(msg);
			}
			catch(PersistenceException e) {
				String msg = "Fehler beim Löschen. " + e.getMessage();
				_debug.warning(msg);
				sendErrorResponse(msg);
			}
		}
	}

	/**
	 * Führt Löschauftrag durch, der von einer Simulation gestellt wurde.
	 *
	 * @param srd Auftragsparameter.
	 *
	 * @throws PersistenceException
	 * @throws IOException
	 */
	private void deleteSimVarSimulation(SimulationResultData srd) throws IOException, PersistenceException {
		boolean successful = false;
		try {
			executeDelete(srd.getSimVar());
			successful = true;
		}
		finally {
			srd.dataDeleted(successful);
		}
	}

	/**
	 * Führt Löschauftrag durch, der von einem Client gestellt wurde.
	 *
	 * @param resultData Auftragsdaten
	 *
	 * @throws PersistenceException
	 * @throws IOException
	 */
	private void deleteSimVarClient(ResultData resultData) throws IOException, PersistenceException {
		try {
			// Task initialisieren.
			init(resultData.getData());
			// SimVar rausholen.
			short simVarToDelete = deserializer.readShort();
			// Löschen durchführen.
			executeDelete(simVarToDelete);
			sendSuccessResponse();
		}
		catch(NoSuchVersionException e) {
			String msg = "Geforderte Version Serialisierers/Deserialisierers nicht verfuegbar. " + e.getMessage();
			_debug.warning(msg);
			sendErrorResponse(msg);
		}
	}

	/**
	 * Löscht die Daten einer Simulationsvariante. Entfernt die betroffenen DataIdentNodes aus dem Baum.<br> Vorgehen: Es müssen alle Verzeichnisse durchlaufen
	 * werden, da es möglich ist, dass Verzeichnisse existieren, die nicht durch DataIdentNodes repräsentiert werden.
	 *
	 * @param simVarToDelete Simulationsvariante.
	 *
	 * @throws IOException          Fehler beim Deserialisieren der Daten aus <code>data</code>.
	 * @throws PersistenceException Fehler beim Löschen der Daten.
	 */
	private void executeDelete(short simVarToDelete) throws IOException, PersistenceException {
		// Achtung: Es ist möglich, dass es hier zu einer kurzzeitigen Inkonsistenz kommt:
		// Zwei Threads greifen auf die gleiche DIN zu, der Löschtask 'gewinnt', der andere wartet
		// auf die Freigabe der DIN. Wenn der die Freigabe erhält wurde das zur DIN gehörenden
		// Verzeichnis bereits gelöscht.

		ContainerFileDirIterator iterator = new ContainerFileDirIterator(persMgr.getArchivePath());
		File simVarDir;
		while((simVarDir = iterator.getNextSimVarDirectory()) != null) {
			suspendTaskIfNecessary();

			long di[] = PersistenceManager.extractDIdFromPath(simVarDir);
			if(di[3] == simVarToDelete) {
				DataIdentNode din = null;
				boolean criticalAreaEntered = false;
				boolean busyAreaEntered = false;
				try {
					// Synchronisierung ueber DataIdentNode:
					din = didTree.existsCreate(di[0], di[1], di[2], (int)di[3]);
					criticalAreaEntered = din.enterCriticalArea();
					busyAreaEntered = din.enterBusyArea();
					if(!criticalAreaEntered || !busyAreaEntered) {
						if(criticalAreaEntered) din.exitCriticalArea();
						if(busyAreaEntered) din.exitBusyArea();
						_debug.warning("Loeschen der Simulationsvariante fehlgeschlagen fuer Verzeichnis " + simVarDir);
						continue; // Dieses SV-Verzeichnis ueberspringen da Zugriff nicht erlaubt wurde.
					}

					try {
						final CacheManager cacheManager = CacheManager.getInstance();
						if(din.hasIndexAssigned()) {
							cacheManager.forgetCache(din.getOpenContID(ArchiveDataKind.ONLINE));
							cacheManager.forgetCache(din.getOpenContID(ArchiveDataKind.ONLINE_DELAYED));
							cacheManager.forgetCache(din.getOpenContID(ArchiveDataKind.REQUESTED));
							cacheManager.forgetCache(din.getOpenContID(ArchiveDataKind.REQUESTED_DELAYED));
						}
					}
					catch(Exception ignored) {
						ignored.printStackTrace();
					}

					// Zuerst SV-Verzeichnis loeschen und danach DIN, da ueber diesen synchronisiert wird:
					persMgr.deleteSimVar(di[0], di[1], di[2], (int)di[3]);
					didTree.deleteDataIdentNode(di[0], di[1], di[2], (int)di[3]);
				}
				catch(DIdNodeNotFoundException e) {
					// Loeschen des DIN fehlgeschlagen
					_debug.warning("Loeschen eines DataIdentNodes fehlgeschlagen fuer DIN " + din, e);
				}
				finally {
					if(din != null && criticalAreaEntered) din.exitCriticalArea();
					if(din != null && busyAreaEntered) din.exitBusyArea();
				}
			}
		}
	}
}
