/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.InQueuesMgr;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTreeIterator;
import de.bsvrz.ars.ars.mgmt.datatree.lock.DinLock;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.ars.ars.persistence.iter.DeletedTreatment;
import de.bsvrz.ars.ars.persistence.iter.QueryDataSequence;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.*;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataSerializer.Deserializer;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.*;
import java.util.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static de.bsvrz.sys.funclib.losb.util.Util.dIdx2StrExt;

/**
 * Task zur Nachforderung fehlender Archivdaten aus anderen Archivsystemen gemaess Parametrierung.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RequestGapTask extends SingleClientSenderTask {

	/** Dateiname der Datenluecken-Datei. */
	public static String GAP_FILE_NAME = "_datagaps.txt";

	/** Anzahl der Threads zum Nachfordern fehlender Archivdaten. */
	public static int REQUEST_GAP_THREADS = 1;

	/** Prioritaet der Nachforserungsanfragen. */
	public static ArchiveQueryPriority REQUEST_PRIO = ArchiveQueryPriority.LOW;

	private static String GAP_KEYWORD = "<not_in>";

	private static String GAP_IDX_SEPARATOR = "-";

	private static String GAP_ARCH_SEPARATOR = "#";

	private static Pattern GAP_PAT = Pattern.compile("^(\\d+)" + GAP_IDX_SEPARATOR + "(\\d+)" + GAP_KEYWORD + "(\\S+)$");

	private static ArchiveDataKindCombination requestAdk = Util.getADKCombination(true, true, true, true);

	private static RequestGapResultData requestGapResultData = null;

	private RequestGapWorker[] requester = null;

	private Semaphore finishedSema;

	private ArchiveDataSpecification[] archDataSpecs;

	private List<SystemObject> archiveList;

	private AtomicInteger totalGapsFound = new AtomicInteger(0);

	/**
	 * Zeitspezifikation, die beim Nachfordern verwendet werden soll oder <code>null</code>, falls keine Einschränkung berücksichtigt werden soll.
	 */
	private ArchiveTimeSpecification archTimeSpec;

	/**
	 * Erzeugt einen neuen Task zur Nachforderung fehlender Archivdaten.
	 *  @param rodl       Der Listener wird benachrichtigt beim Beenden der Task.
	 * @param archiveMgr Der Archiv-Manager.
	 * @param inputData  Eingangswarteschlange
	 */
	public RequestGapTask(ArchiveManager archiveMgr, BlockingQueue<Object> inputData) {
		super(archiveMgr, inputData);
	}

	/** Liefert den Nachforderungsanfrage-Datensatz. */
	public static RequestGapResultData getRequestGapResultData() {
		if(requestGapResultData == null) requestGapResultData = new RequestGapResultData();
		return requestGapResultData;
	}

	/**
	 * Startet das automatische oder manuelle Nachfordern entsprechend dem uebergebenen Nachforderungs-Datensatz.
	 *
	 * @param resultData Nachforderungs-Datensatz.
	 */
	protected void work(ResultData resultData) {
		long t = Util.startTimer();
		totalGapsFound.set(0);
		long anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested();
		if(resultData instanceof RequestGapResultData) {
			_debug.info(getName() + ": Automatische Nachforderung gestartet.");
			archTimeSpec = null;
			archDataSpecs = new ArchiveDataSpecification[0];
			archiveList = new ArrayList<SystemObject>(0);
			try {
				performRequest();
				anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested() - anzReqDS;
				_debug.info(
						getName() + ": Automatische Nachforderung beendet (Zeit: " + Util.stopTimer(t) + ", Luecken: " + totalGapsFound.get()
								+ ", erhaltene Datensaetze: " + anzReqDS + ")"
				);
			}
			catch(Exception e) {
				_debug.warning("Nachforderung fehlgeschlagen.", e);
			}
		}
		else {
			if(!resultData.hasData()) return;
			int msgType = resultData.getData().getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).intValue();
			if(msgType == MessageType.REQUEST_DID) {
				int messageTypeResult = MessageType.REQUEST_DID_RESULT;
				try {
					_debug.info(getName() + ": Manuelle Nachforderung gestartet.");
					init(resultData.getData());
					archTimeSpec = null;
					archDataSpecs = parseArchiveDataSpec(deserializer, resultData);
					archiveList = parseArchiveList(deserializer);
					performRequest();
					sendSuccessResponse(messageTypeResult);
					anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested() - anzReqDS;
					_debug.info(
							getName() + ": Manuelle Nachforderung beendet (Zeit: " + Util.stopTimer(t) + ", Luecken: " + totalGapsFound.get()
									+ ", erhaltene Datensaetze: " + anzReqDS + ")"
					);
				}
				catch(Exception e) {
					_debug.warning("Nachforderung fehlgeschlagen.", e);
					sendErrorResponse(messageTypeResult, "Nachforderung fehlgeschlagen: " + e.getMessage());
				}
			}
			else if(msgType == MessageType.REQUEST_AUTOM) {
				int messageTypeResult = MessageType.REQUEST_AUTOM_RESULT;
				try {
					_debug.info(getName() + ": Manueller Anstoß der automatischen Nachforderung.");
					init(resultData.getData());
					archDataSpecs = new ArchiveDataSpecification[0];
					final long startTime = deserializer.readLong();
					final long endTime = deserializer.readLong();
					archTimeSpec = new ArchiveTimeSpecification(TimingType.ARCHIVE_TIME, false, startTime, endTime);
					archiveList = parseArchiveList(deserializer);
					performRequest();
					sendSuccessResponse(messageTypeResult);
					anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested() - anzReqDS;
					_debug.info(
							getName() + ": Manueller Anstoß der automatischen Nachforderung beendet (Zeit: " + Util.stopTimer(t) + ", Luecken: " + totalGapsFound
									.get()
									+ ", erhaltene Datensaetze: " + anzReqDS + ")"
					);
				}
				catch(Exception e) {
					_debug.warning("Nachforderung fehlgeschlagen.", e);
					sendErrorResponse(messageTypeResult, "Manueller Anstoß der automatischen Nachforderung fehlgeschlagen: " + e.getMessage());
				}
			}
			else {
				_debug.error("Ungültiger Nachrichtentyp für die Nachforderung: " + msgType);
			}
		}
		archDataSpecs = null;
		requester = null;
	}

	/**
	 * Extrahiert die Liste der Archivsysteme, bei denen nachgefordert werden soll. Der Deserialisierer muss genau vor dem Laengenbyte der Liste stehen.
	 * Doppelte EIntraege werden gefiltert, Nachfordern bei sich selbst ebenso.
	 *
	 * @param ds
	 * @return Liste mit Archivsystem-Objekten, leere Liste, falls keine angegeben sind
	 * @throws IOException
	 */
	private List<SystemObject> parseArchiveList(Deserializer ds) throws Exception {
		int anzArchives = deserializer.readInt(); // Größe der Liste mit den Archiven
		List<SystemObject> result = new ArrayList<SystemObject>(anzArchives);

		HashSet<SystemObject> singleEntries = new HashSet<SystemObject>();
		for(int i = 0; i < anzArchives; i++) {
			SystemObject foreignArS = ds.readObjectReference(archMgr.getDataModel());
			if(foreignArS == null) throw new Exception("Archivsystem-Objekt[" + i + "] == null");

			// doppelte Anfragen vermeiden, nicht bei sich selber nachfragen
			if(!singleEntries.contains(foreignArS) && !archMgr.getConfigAuth().getPid().equals(foreignArS.getPid())) {
				result.add(foreignArS);
				singleEntries.add(foreignArS);
			}
		}
		return result;
	}

	/**
	 * Erzeugt eine Anzahl ({@link #REQUEST_GAP_THREADS}) von {@link #RequestGapTask(ArchiveManager, RingOfDeathListener, RingBuffer)}- Objekten, die die
	 * Nachforderung durchführen. Diese Methode wartet, bis alle Threads fertig sind.
	 *
	 * @throws Exception
	 */
	private void performRequest() throws Exception {
		for(ArchiveDataSpecification ads : archDataSpecs) {
			if(ads.getTimeSpec().isStartRelative()) throw new Exception("Relative Zeit-/Indexspezifikation nicht moeglich beim Nachfordern");
		}

		finishedSema = new Semaphore(REQUEST_GAP_THREADS);
		try {
			finishedSema.acquire(REQUEST_GAP_THREADS);
			requester = new RequestGapWorker[REQUEST_GAP_THREADS];
			for(int i = 0; i < REQUEST_GAP_THREADS; i++) {
				requester[i] = new RequestGapWorker(archMgr, i);
				requester[i].start();
			}
			finishedSema.acquire(REQUEST_GAP_THREADS);
			requester = null;
		}
		catch(InterruptedException e) {
			_debug.info(getName() + ": Nachforderungslauf wurde abgebrochen");
		}
	}

	/** Beendet den Nachforderungs-Task. */
	public void terminateTask() {
		if(requester != null) {
			for(RequestGapWorker reqWorker : requester) {
				if(reqWorker != null) reqWorker.terminateTask();
			}
		}
		super.terminateTask();
		requester = null;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Klasse RequestGapWorker
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Speichert pro Archivsystem die Luecken, die von diesem nicht geschlossen werden konnten. Am Schluss wird die gesamte Lueckenliste aller Archivsysteme
	 * durchgegangen und so zusammengefasst, dass nur die Luecken erhalten bleiben, die von keinem Archivsystem geschlossen werden konnten.
	 */
	public static class GapMarker {

		private Map<String, List<Long>> gaps = new HashMap<String, List<Long>>();

		// Für den Durchlauf: merkt sich, welches ArS bei welcher Luecke steht
		private Map<String, Integer> pos = new HashMap<String, Integer>();

		public void reset() {
			gaps.clear();
		}

		public void addGap(String archive, long gapStart, long gapEnd) {
			if(!gaps.containsKey(archive)) gaps.put(archive, new ArrayList<Long>());

			List<Long> ls = gaps.get(archive);
			if(ls.size() >= 2) {
				Long lastGapStart = ls.get(ls.size() - 2);
				Long lastGapEnd = ls.get(ls.size() - 1);
				if(lastGapStart.equals(gapStart) && lastGapEnd.equals(gapEnd))
				// keine doppelten Luecken hinzufuegen!!
				{
					return;
				}
			}
			gaps.get(archive).add(gapStart);
			gaps.get(archive).add(gapEnd);
		}

		public void addGap(String[] parentFailedArS, String archive, long gapStart, long gapEnd) {
			addGap(archive, gapStart, gapEnd);
			for(String arch : parentFailedArS) {
				addGap(arch, gapStart, gapEnd);
			}
		}

		/**
		 * Geht die gesamte Lueckenliste aller Archivsysteme durch und fasst sie so zusammen, dass nur die Luecken erhalten bleiben, die von keinem Archivsystem
		 * geschlossen werden konnten. Diese Liste wird in Zeilen der Form 100-200<not in>archiv1#archiv2 ausgegeben.
		 */
		public String reduce() {
			StringBuilder sb = new StringBuilder(2048);
			String archives = "";
			pos.clear();
			int i = 0;
			List<String> archiveNames = new ArrayList<String>(gaps.keySet());
			Collections.sort(archiveNames);
			for(String arch : archiveNames) {
				pos.put(arch, 0);
				archives += arch + (i++ < gaps.size() - 1 ? GAP_ARCH_SEPARATOR : "");
			}

			Long fullGapStart = null, fullGapEnd = null;
			long lastDidx = -1;
			String archMarker;

			while((archMarker = getNext()) != null) {
				long didx = getDIdx(archMarker);
				if(didx != lastDidx) {
					if(isFullGap(didx)) {
						fullGapEnd = (fullGapStart != null) ? new Long(didx) : null;
						fullGapStart = (fullGapStart == null) ? new Long(didx) : fullGapStart;
						if(fullGapStart != null && fullGapEnd != null) {
							sb.append(fullGapStart + GAP_IDX_SEPARATOR + fullGapEnd + GAP_KEYWORD + archives + "\n");
							fullGapStart = null;
							fullGapEnd = null;
						}
					}
					lastDidx = didx;
				}
				pos.put(archMarker, pos.get(archMarker) + 1);
			}
			// Luecken werden oben immer nur beim naechsten Durchlauf ausgegeben. Drum hier pruefen,
			// ob die Liste mit einer Luecke endet:
			if(fullGapStart != null && fullGapEnd != null) sb.append(fullGapStart + GAP_IDX_SEPARATOR + fullGapEnd + GAP_KEYWORD + archives + "\n");
			return sb.toString();
		}

		/**
		 * @param didx
		 * @return Wahr, wenn der Datenindex von allen Archivsystemen nicht geliefert werden konnte.
		 */
		private boolean isFullGap(long didx) {
			for(String arch : gaps.keySet()) {
				List<Long> marks = gaps.get(arch);
				boolean isGap = false;
				for(int i = 1; i < marks.size(); i += 2) {
					if(marks.get(i - 1) <= didx && didx <= marks.get(i)) {
						isGap = true;
						break;
					}
				}
				if(!isGap) return false;
			}
			return true;
		}

		/** @return Das Archivsystem, das den naechstkleinsten Datenindex vermerkt hat */
		private String getNext() {
			String minArch = null;
			for(String arch : gaps.keySet()) {
				if(pos.get(arch) < gaps.get(arch).size()) if(minArch == null || getDIdx(arch) < getDIdx(minArch)) minArch = arch;
			}
			return minArch;
		}

		private long getDIdx(String archive) {
			return gaps.get(archive).get(pos.get(archive));
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Klasse GapMarker
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Diese Klasse sieht aus wie ein ResultData, wird aber nur intern verwendet, um einen über den Scheduler angestoßenen Sicherungslauf von einem extern
	 * angestoßenen zu unterscheieden
	 *
	 * @author beck et al. projects GmbH
	 * @author Phil Schrettenbrunner
	 */
	public static class RequestGapResultData extends ResultData {

		/** Erzeugt ein neues leeres RequestGapResultData-Objekt. */
		public RequestGapResultData() {
			super(null, null, 0, (Data) null);
		}

	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Klasse GapFile
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	private class RequestGapWorker extends Task {

		public int workerIndex;

		private DataDescription dataDesc;

		private SystemObject object;

		private InQueuesMgr iqm = archMgr.getInQueuesMgr();

		private byte[] indexMem = new byte[INDEX_MEMORY_SIZE];

		public RequestGapWorker(ArchiveManager aMgr, int wkIndex) {
			super(aMgr);
			workerIndex = wkIndex;
			setName(getName() + "_" + workerIndex);
			iqm = archMgr.getInQueuesMgr();
		}

		@Override
		public void execute() {
			try {
				if(archDataSpecs.length == 0) {
					// Keine Datenspezifikation, also alles im DataIdentTree nachfordern
					// Da Nachfordern ueber die Parametrierung spezifiziert wird, genuegt es, ueber
					// den DataIdentTree zu laufen.

					DataIdentTreeIterator it = didTree.createIterator();
					while(it.getNext()) {
						if(shouldTerminate()) break;
						if((it.getObjId() + it.getAtgId() + it.getAspId() + it.getSV()) % REQUEST_GAP_THREADS == workerIndex) {
							requestDID(it.getObjId(), it.getAtgId(), it.getAspId(), (short) it.getSV(), it.getDataIdentNode(), archTimeSpec);
						}
					}
				}
				else {
					// nur die in der Datenspezifikation genannten DIDs nachfordern
					for(ArchiveDataSpecification ads : archDataSpecs) {
						if(shouldTerminate()) break;
						long obj = ads.getObject().getId();
						long atg = ads.getDataDescription().getAttributeGroup().getId();
						long asp = ads.getDataDescription().getAspect().getId();
						short sv = (short) ads.getDataDescription().getSimulationVariant();

						if((obj + atg + asp + sv) % REQUEST_GAP_THREADS == workerIndex) {
							try {
								requestDID(obj, atg, asp, sv, didTree.getDataIdentNode(ads), ads.getTimeSpec());
							}
							catch(DIdNodeNotFoundException e) {
								// Es wird kein DIN eingefuegt, weil er parametriert sein (wegen Nachforderungs-Archiven)
								// und deshalb existieren muss.
							}
						}
					}
				}
			}
			finally {
				persMgr.removeContainerFileObjects(this);
				finishedSema.release();
				super.execute();
			}
		}

		/**
		 * Fordert die angegebene Datenidentifikation innerhalb des angegebenen Bereichs nach. Diese Methode pausiert, solange das Archivsystem im kritsichen
		 * Zustand (d.h. Warteschlange fast voll) ist.
		 */
		private void requestDID(long obj, long atg, long asp, short sv, DataIdentNode din, ArchiveTimeSpecification ats) {
			GapMarker newGaps = new GapMarker();
			GapFile gapFile = null;
			boolean busyAreaEntered;
			List<SystemObject> requestArchives;

			try {
				
				suspendTaskIfNecessary();    //Hält den Task an, falls das Archivsystem im kritischen Bereich.
				
				if(!din.arSParamIsNachfordern()) return;    // Nachfordern nicht parametriert

				// Synchronisierung ueber DataIdentNode (Busy-Area):
				try(DinLock lock = din.lockBusy()) {
					if(lock == null) {
						_debug.warning("Nachfordern fehlgeschlagen fuer Datenidentifikation: " + din);
						return;
					}

					dataDesc = new DataDescription(archMgr.getAtg(atg), archMgr.getAsp(asp));
					object = archMgr.getObj(obj);

					// Wenn eine Archivliste gegeben ist, diese verwenden, sonst Parametrierung
					requestArchives = archiveList.size() > 0 ? archiveList : copyReqArSFromDIN(din);

					QueryDataSequence dataSequence = new QueryDataSequence(archMgr, new IdDataIdentification(obj, atg, asp, sv), ArchiveDataKindCombination
							.all(), ats, ArchiveOrder.BY_INDEX, DeletedTreatment.Full, this, indexMem);

					boolean skipLine = false;
					try(DataIterator iterator = dataSequence.iterator()) {

						long lowDidx = -1, lowDTime = -1, lowATime = -1;

						gapFile = new GapFile(obj, atg, asp, sv);
						gapFile.createTmpGapFile();
						boolean startCondition = false;
						while(!iterator.isEmpty()) {

							if(shouldTerminate()) break;

							long hiDidx = iterator.peekDataIndex();

							// Pruefen, ob wir im angegebenen Bereich sind. Dies ist der Fall, wenn die Luecke
							// vollstaendig im Bereich liegt. Ggf. Lueckendatei vor und nach dem Bereich kopieren.
							if(!startCondition) {
								if(startCondition(ats, lowDidx, lowDTime, lowATime)) {
									gapFile.copyGapsUpToHere(lowDidx);
									startCondition = true;
								}
							}
							if(startCondition) {

								if(isGap(lowDidx, hiDidx)) {
									totalGapsFound.incrementAndGet();
									newGaps.reset();
									// rausfinden, wo wir fragen muessen, wo erfolglos gefragt wurde und wo von den erfolglosen
									// wir nicht mehr fragen (weil die auch weiterhin in der Lueckendatei vermerkt sein sollen):
									String[] failedArchives = gapFile.getEntry(lowDidx, hiDidx);
									String[] unaskedArchives = extractUnaskedArS(requestArchives, failedArchives);

									skipLine = failedArchives.length > 0;    // wenn es einen Eintrag gab, diesen beim Kopieren ueberspringen
									requestGap(requestArchives, 0, lowDidx, hiDidx, unaskedArchives, gapFile, newGaps);
									gapFile.appendToTmpGapFile(newGaps.reduce());
								}
							}
							lowDidx = hiDidx;
							lowDTime = iterator.peekDataTime();
							lowATime = iterator.peekArchiveTime();
							iterator.remove();
						}
					}
					gapFile.copyRestGaps(skipLine);
				}
			}
			catch(InterruptedException e) {
				_debug.fine("Nachfordern wurde unterbrochen: '" + Util.did2Str(obj, atg, asp, sv) + "': " + e.getMessage());
			}
			catch(Exception e) {
				_debug.warning("Fehler beim Nachfordern von '" + Util.did2Str(obj, atg, asp, sv), e);
			}
			finally {
				if(gapFile != null) gapFile.activateTmpGapFileClose();
			}
		}

		/**
		 * Prueft, ob der anstehende Datensatz innerhalb der Grenzen liegt
		 *
		 * @param ats      Grenzen
		 * @param lowDidx  Datenindex der unteren Lueckenkante
		 * @param lowDTime Datenzeit der unteren Lueckenkante
		 * @param lowATime Archivzeit der unteren Lueckenkante
		 * @return Wahr, wenn der anstehende Datensatz innerhalb der Grenzen liegt, falsch sonst
		 */
		private boolean startCondition(ArchiveTimeSpecification ats, long lowDidx, long lowDTime, long lowATime) {
			if(ats == null) {
				return true;
			}
			else if(ats.getTimingType().equals(TimingType.DATA_INDEX)) {
				return lowDidx > -1 && lowDidx >= ats.getIntervalStart();
			}
			else if(ats.getTimingType().equals(TimingType.DATA_TIME)) {
				return lowDTime > -1 && lowDTime >= ats.getIntervalStart();
			}
			else if(ats.getTimingType().equals(TimingType.ARCHIVE_TIME)) {
				return lowATime > -1 && lowATime >= ats.getIntervalStart();
			}
			else {
				return false;
			}
		}

		private String[] extractUnaskedArS(List<SystemObject> reqArchives, String[] failedArchives) {
			List<String> temp = new ArrayList<String>();
			for(String arch : failedArchives) {
				if(!Util.containsPid(reqArchives, arch)) temp.add(arch);
			}
			String[] erg = new String[temp.size()];
			for(int i = 0; i < erg.length; i++) {
				erg[i] = temp.get(i);
			}
			return erg;
		}

		/**
		 * Kopiert die Liste der Nachforderungs-Archive aus dem DataIdentNode. Jedes Archivsystem wird nur einmal in die Ergebnisliste aufgenommen.
		 *
		 * @param din
		 * @return Liste von Archiv-Objekten
		 */
		private List<SystemObject> copyReqArSFromDIN(DataIdentNode din) {
			ArrayList<SystemObject> otherArchives = new ArrayList<SystemObject>();
			HashSet<String> singleEntries = new HashSet<String>();
			for(int i = 0; i < din.arSParamGetAnzNachfordern(); i++) {
				String foreignArS = din.arSParamGetNachfordern(i);

				if(foreignArS.equals("0")) continue;    // DAV erlaubt hier Nullwerte

				// doppelte Anfragen vermeiden, nicht bei sich selber nachfragen
				if(!singleEntries.contains(foreignArS) && !archMgr.getConfigAuth().getPid().equals(foreignArS)) {
					otherArchives.add(archMgr.getObj(din.arSParamGetNachfordern(i)));
					singleEntries.add(din.arSParamGetNachfordern(i));
				}
			}
			return otherArchives;
		}

		/**
		 * Diese Methode fragt die uebergebene Datenluecke rekursiv bei der Liste der Archivsysteme ab. Bei jeder Luecke in der aktuellen Anfrage wird rekursiv
		 * eine Anfrage bis Lueckenende an das in der Liste nachfolgende Archivsystem gestellt. Dadurch werden immer <b>alle</b> parametrierten Archivsysteme
		 * angefragt, bevor der Datenindex an ein Lueckenende springt. Wenn für eine Luecke und ein Archivsystem vermerkt ist, dass dieses die Luecke nicht
		 * schliessen kann, wird keine Nachforderungsanfrage gestellt.
		 *
		 * @param archs  Liste von Archivsystemen
		 * @param aIdx   Archivsystem in der Liste, bei dem diese Anfrage gestellt werden soll
		 * @param lowBnd Untergrenze der Anfrage (letzter vorhandener DS)
		 * @param upBnd  Obergrenze der Anfrage (erster wieder vorhandener DS)
		 * @throws InterruptedException
		 */
		private void requestGap(
				List<SystemObject> archs, int aIdx, long lowBnd, long upBnd, String[] unaskedPreviousFailedArS, GapFile oldGaps, GapMarker newGaps)
				throws InterruptedException {


			if(!isGap(lowBnd, upBnd)) return;

			// Wenn keine ArS zum Nachfordern parametriert sind, werden die alten Lueckeninfos uebernommen
			if(archs.size() == 0) {
				for(String arch : unaskedPreviousFailedArS) {
					newGaps.addGap(arch, lowBnd, upBnd);
				}
				return;
			}

			// Am Ende der Liste der Nachfordern-Archivsysteme angelangt?
			if(aIdx >= archs.size()) return;

			long nextExpNoModBits = noModBits(lowBnd) + 1;    // der naechste erwartete Index (ohne Ars-/DaV-Bit)
			long lastSuccess = lowBnd;                        // der letzte vorhandene Index

			String[] failedArS = oldGaps.getEntry(lowBnd, upBnd);

			// Wenn Archivsysteme angegeben sind, wird dort auch dann nachgefragt, wenn dies
			// bereits als nicht erfolgreich vermerkt wurde (TAnfArS 5.1.2.12.2)
			if(archiveList.size() > 0 || !Util.containsStr(failedArS, archs.get(aIdx).getPid())) {
				_debug.fine(
						"Nachforderungsanfrage an '" + archs.get(aIdx) + "': startIdx=" + lowBnd + " (" + dIdx2StrExt(lowBnd) + "), endIdx=" + upBnd + " ("
								+ dIdx2StrExt(upBnd) + ")"
				);

				ArchiveRequestManager arm = archMgr.getDavCon().getArchive(archs.get(aIdx));
				// Anm.: Die Methode arm.isArchiveAvailable() liefert bei sofortigem Aufruf immer false und wurde ausgelassen

				try {
					ArchiveDataQueryResult res = arm.request(REQUEST_PRIO, buildRequest(lowBnd, upBnd));
					if(res.isRequestSuccessful()) {
						ArchiveDataStream ads = res.getStreams()[0];    // nur einer war angelegt
						ArchiveData ad;
						try {
							while((ad = ads.take()) != null) {
								if(idx(ad) >= upBnd) break;         // drueber (letzter)

								if(idx(ad) <= lowBnd || noModBits(ad) < nextExpNoModBits) {
									continue;    // drunter, erster oder potGap mit gleichem Index wie letzter DS
								}

								if(noModBits(ad) > nextExpNoModBits) {
									// Luecke in der Antwort! Der erhaltene DIdx ist Größer als der erwartete.
									// Die neu entstandene Luecke wird rekursiv bei den anderen Archivsystemen angefragt.
									newGaps.addGap(unaskedPreviousFailedArS, archs.get(aIdx).getPid(), lastSuccess, idx(ad));
									requestGap(
											archs, aIdx + 1, lastSuccess, idx(ad), unaskedPreviousFailedArS, oldGaps, newGaps
									);    // mit allen ArS versuchen zu fuellen
									nextExpNoModBits = noModBits(ad);                    // weiter, egal ob fuellen erfolgreich war oder nicht
								}
								// Datenindex passt. Nur speichern, wenn korrekte Datenart. Lueckenmarkierungen
								// die einen fehlerhaften Index haben, werden dadurch nicht gespeichert.
								if(properDataState(ad)) iqm.insertInArchiveQueueReq(System.currentTimeMillis(), ad);

								lastSuccess = idx(ad);
								nextExpNoModBits++;
							}
						}
						catch(IOException e) {
							_debug.warning("Fehler beim Empfang eines Nachforderungs-Datensatzes von '" + archs.get(aIdx) + "'", e);
						}
						finally {
							ads.abort();
						}
					}
					else {
						_debug.warning("Nachforderungsanfrage an Archivsystem '" + archs.get(aIdx) + "' fehlgeschlagen", res.getErrorMessage());
					}
				}
				catch(IllegalStateException e) {
					_debug.fine("Nachfordern: Archivsystem '" + archs.get(aIdx) + "' ist nicht erreichbar");
				}
			}
			else {
				_debug.finest(
						"Nachforderungsanfrage an '" + archs.get(aIdx) + "' bereits als nicht erfolgreich vermerkt: startIdx=" + lowBnd + " ("
								+ dIdx2StrExt(lowBnd) + "), endIdx=" + upBnd + " (" + dIdx2StrExt(upBnd) + ")"
				);
			}

			// Haben wir auch nach der Antwort noch eine Luecke oder hat das angefragte ArS gar nichts geliefert?
			if(isGap(lastSuccess, upBnd)) {
				newGaps.addGap(unaskedPreviousFailedArS, archs.get(aIdx).getPid(), lastSuccess, upBnd);
				// naechstes Archivsystem fragen:
				requestGap(archs, aIdx + 1, lastSuccess, upBnd, unaskedPreviousFailedArS, oldGaps, newGaps);
			}
		}

		// ***************************************************************************************************
		// **********  Abkuerzende Methoden (um den Hauptalgorithmus einfach zu halten) **********************
		// ***************************************************************************************************

		private ArchiveDataSpecification buildRequest(long startIdx, long endIdx) {
			return new ArchiveDataSpecification(
					new ArchiveTimeSpecification(TimingType.DATA_INDEX, false, startIdx, endIdx),
					requestAdk,
					ArchiveOrder.BY_INDEX,
					ArchiveRequestOption.NORMAL,
					dataDesc,
					object
			);
		}

		private boolean isGap(long start, long end) {
			return start == -1 ? false : noModBits(end) - noModBits(start) > 1;
		}

		private boolean properDataState(ArchiveData ad) {
			DataState ds = ad.getDataType();
			return ds.equals(DataState.DATA) || ds.equals(DataState.NO_DATA) || ds.equals(DataState.NO_SOURCE);
		}

		private long idx(ArchiveData ad) {
			return ad.getDataIndex();
		}

		private long noModBits(long dataIndex) {
			return Util.dIdxNoModBits(dataIndex);
		}

		private long noModBits(ArchiveData ad) {
			return Util.dIdxNoModBits(ad.getDataIndex());
		}

	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Klasse RequestGapResultData
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Diese Klasse speichert nicht schliessbare Datenluecken in einer Datei im Verzeichnis der Datenidentifikation in der Form <ul>
	 * <li>10-20<not_in>archiv1#archiv2 <li>30-40<not_in>archiv3#archiv4 <li>... </ul> Für jede Luecke kann dann der entsprechende Eintrag abgefragt werden.
	 */
	public class GapFile {

		private File gapFile, tmpGapFile;

		private BufferedReader input = null;

		private BufferedWriter output = null;

		public GapFile(long objID, long atgID, long aspID, short sv) {
			this(persMgr.getPathName(objID, atgID, aspID, sv));
		}

		public GapFile(String dir) {
			gapFile = new File(dir, GAP_FILE_NAME);
			try {
				input = new BufferedReader(new InputStreamReader(new FileInputStream(gapFile)));
			}
			catch(FileNotFoundException e) {
			}    // es muss nicht unbedingt eine geben
		}

		/**
		 * Legt eine temporaere Lueckendatei an, in der die sich neu ergebenden Luecken gesammelt werden. Am Ende eines Nachforderungslaufs wird die Datei
		 * umbenannt.
		 */
		public void createTmpGapFile() {
			tmpGapFile = new File(gapFile.getAbsolutePath() + ".tmp");
			try {
				Util.deleteCreateNewFile(tmpGapFile);
				output = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(tmpGapFile)));
			}
			catch(IOException e) {
				_debug.warning("Temporaere Lueckendatei '" + tmpGapFile.getAbsolutePath() + " konnte nicht angelegt werden: " + e.getMessage());
			}
		}

		/**
		 * Uebertraegt alle Eintraege bis zur genannten Grenze in die temporaere Lueckendatei. Dies ist notwendig, wenn nur bestimmte Bereiche nachgefordert
		 * werden sollen.
		 *
		 * @param startGap Untergrenze des Bereichs, der nachgefordert werden soll
		 */
		public void copyGapsUpToHere(long startGap) {
			if(input == null) return;
			try {
				while(true) {
					input.mark(4096);
					String line = input.readLine();
					if(line == null) break;
					Matcher m = GAP_PAT.matcher(line);
					if(m.matches()) {
						long entryStart = Long.parseLong(m.group(1));
						if(entryStart < startGap) {
							appendToTmpGapFile(line + "\n");
						}
						else {
							input.reset();
							break;
						}
					}
				}
			}
			catch(IOException e) {
				_debug.warning("Lueckendatei '" + gapFile.getAbsolutePath() + " konnte nicht gelesen werden: " + e.getMessage());
			}
		}

		/**
		 * Kopiert die Eintraege hinter der aktuellen Position in die temporaere Lueckendatei. Dies ist notwendig, wenn nur bestimmte Bereiche nachgefordert
		 * werden sollen. Wenn der naechste Eintrag exakt die Luecke ist, die nachgefordert wurde, muss diese Zeile uebersprungen werden (weil dieser Eintrag
		 * ueber den GapMarker eingefuegt wurde -- oder auch nicht, je nach Erfolg der Nachforderung).
		 *
		 * @param skipLine
		 */
		public void copyRestGaps(boolean skipLine) {
			if(input == null) return;
			try {
				String line = skipLine ? input.readLine() : null;    // ggf. Zeile mit Lueckeneintrag ueberspringen
				while((line = input.readLine()) != null) {
					Matcher m = GAP_PAT.matcher(line);
					if(m.matches()) appendToTmpGapFile(line + "\n");
				}
			}
			catch(IOException e) {
				_debug.warning("Lueckendatei '" + gapFile.getAbsolutePath() + " konnte nicht gelesen werden: " + e.getMessage());
			}
		}

		/**
		 * Fuegt der Lueckendatei einen Eintrag hinzu. Der Eintrag muss bereits korrekt formatiert sein.
		 *
		 * @param entry Lueckeneintrag
		 */
		public void appendToTmpGapFile(String entry) {
			try {
				output.write(entry);
			}
			catch(IOException e) {
				_debug.warning("Temporaere Lueckendatei '" + tmpGapFile.getAbsolutePath() + " konnte nicht beschrieben werden: " + e.getMessage());
			}
		}

		/** Loescht die urspruengliche Lueckendatei, benennt die temporaere Lueckendatei in eine echte Luecjendatei um und schliesst alles. */
		public void activateTmpGapFileClose() {
			if(input != null) {
				try {
					input.close();
				}
				catch(IOException e) {
				}
				gapFile.delete();
			}
			try {
				if(output != null) output.close();
				tmpGapFile.renameTo(gapFile);
			}
			catch(IOException e) {
				_debug.warning("Lueckendatei '" + gapFile.getAbsolutePath() + " konnte nicht angelegt werden: " + e.getMessage());
			}
		}

		/**
		 * Liefert die Archivsysteme, die für die angegebene Luecke eingetragen sind
		 *
		 * @param reqGapStart Beginn der Luecke
		 * @param reqGapEnd   Ende der Luecke
		 * @return Feld von Archivsystemen falls ein Eintrag existiert, leeres Feld sonst
		 */
		public String[] getEntry(long reqGapStart, long reqGapEnd) {
			if(input == null) return new String[0];
			try {
				while(true) {
					input.mark(4096);
					String line = input.readLine();
					if(line == null) return new String[0];
					Matcher m = GAP_PAT.matcher(line);
					if(m.matches()) {
						long start = Long.parseLong(m.group(1));
						if(start < reqGapStart) continue;
						if(start > reqGapStart) {
							input.reset();
							return new String[0];
						}
						input.reset();    // Mehrfachanfragen bei gleicher Startzeit sind moeglich
						if(reqGapEnd == Long.parseLong(m.group(2))) {
							return m.group(3).split(GAP_ARCH_SEPARATOR);
						}
						else {
							return new String[0];
						}
					}
					else {
						return new String[0];    // defekte Zeile ueberspringen
					}
				}
			}
			catch(Exception e) {
				_debug.warning("Fehler beim Lesen der Lueckendatei '" + gapFile.getAbsolutePath() + "': " + e.getMessage());
				return new String[0];
			}
		}

		public void close() {
			if(input != null) {
				try {
					input.close();
				}
				catch(IOException e) {
					_debug.warning("Fehler beim Schliessen der Lueckendatei '" + gapFile.getAbsolutePath() + "': " + e.getMessage());
				}
			}
			if(output != null) {
				try {
					output.close();
				}
				catch(IOException e) {
					_debug.warning("Fehler beim Schliessen der Lueckendatei '" + tmpGapFile.getAbsolutePath() + "': " + e.getMessage());
				}
			}
		}
	}

}
