/*
 *
 * Copyright 2009 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

import java.io.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;

/**
 * Diese Klasse implementiert die Verwaltung der Caches für die Schreibzugriffe auf Containerdateien. Die Klasse ist als Singleton realisiert, weil es im
 * Original-Archivsystem keine zentrale Stelle zum Zugriff auf Container-Dateien gibt (der PersistenzManager war wohl dafür vorgesehen, allerdings gibt es auch
 * einige Zugriffe auf Containerdateien, die am PersistenzManager vorbei gehen). Vor dem Zugriff auf Containerdateien muss die Cache-Verwaltung mit der Methode
 * <code>init()</code> erzeugt und initialisiert werden. Beim Beenden des Archivsystems muss mit einem Aufruf der Methode <code>close</code> das Schreiben der
 * noch in den einzelnen Caches vorhandenen Daten veranlasst werden.
 *
 * @author Kappich Systemberatung
 * @version $Revision$
 */
public class CacheManager {

	/** Minimale Puffergröße je Datenidentifikation als Anzahl Bytes. */
	private final static int MINIMUM_BUFFER_SIZE = 0;

	/**
	 * Standard-Puffergröße je Datenidentifikation als Anzahl Bytes. Wird verwendet, wenn kein anderer Wert über das Aufrufargument cachePufferGröße=...
	 * angegeben wurde.
	 */
	private static final int DEFAULT_BUFFER_SIZE = 5 * 1024;

	/** Maximale Puffergröße je Datenidentifikation als Anzahl Bytes. */
	private final static int MAXIMUM_BUFFER_SIZE = 33 * 1024;

	/** Zum aktivieren/deaktivieren von Testausgaben. */
	private final static boolean DEBUG_CALLS = false;

	/** Singleton Objekt des CacheManagers. */
	private static CacheManager _cacheManager = new CacheManager();

	/** Hashmap zur Zuordnung einer Container-ID zum zugehörigen Cache-Objekt. */
	private HashMap<Long, Cache> _containerId2Cache = new HashMap<Long, Cache>();

	/**
	 * Beim Anlegen der Cache-Objekte zu verwendende Puffergröße als Anzahl Bytes. Der Wert kann über das Aufrufargument -cachePufferGröße=... eingestellt
	 * werden.
	 *
	 * @see #DEFAULT_BUFFER_SIZE
	 * @see #init(de.bsvrz.sys.funclib.commandLineArgs.ArgumentList)
	 */
	private int _defaultBufferSize = DEFAULT_BUFFER_SIZE;

	/**
	 * Hashmap mit den Datenidentifikationen, die nicht gepuffert, sondern sofort persistiert werden. Die Hashmap wird von der Methode {@link
	 * #setCachingEnabled(long,long,long,int,boolean)} modifiziert und dient dazu, die Zwischenspeicherung von Datenidentifikationen, bei denen der
	 * Quittierungsmechanismus des Archivsystems verwendet wird, zu verhindern.
	 */
	private HashSet<IdDataIdentification> _disabledCacheIdentifications = new HashSet<IdDataIdentification>();

	/** Debug-Logger für Debug-Ausgaben. */
	private Debug _debug;

	/** Gesamtanzahl von Datei-Schreibvorgängen aller geschlossenen Cache-Objekte. */
	private long _writeCountSum = 0;

	/** Gesamtanzahl von verarbeiteten Datenblöcken aller geschlossenen Cache-Objekte. */
	private long _processedCountSum = 0;

	/** Gesamtanzahl von zwischengespeicherten Datenblöcken aller geschlossenen Cache-Objekte. */
	private long _bufferedCountSum = 0;

	/** Gesamtanzahl von nicht zwischengespeicherten Datenblöcken aller geschlossenen Cache-Objekte. */
	private long _unbufferedCountSum = 0;

	/** Anzahl der geschriebenen Bytes */
	private long _writtenBytes = 0;

	private boolean _cacheEnabled = true;

	/** Erzeugt eine neue Cacheverwaltung. */
	private CacheManager() {
	}

	/** @return Liefert das CacheManager-Objekt zurück. */
	public static CacheManager getInstance() {
		return _cacheManager;
	}

	/**
	 * Initialisiert den CacheManager. Das Aufrufargument -cachePufferGröße=..., mit dem die Größe des Puffers pro Cache eingestellt werden kann, wird gelesen.
	 *
	 * @param argumentList Aufrufargumente der Applikation
	 */
	public void init(final ArgumentList argumentList) {
		if(DEBUG_CALLS) System.out.println("CacheManager.init " + argumentList);
		_writeCountSum = 0;
		_processedCountSum = 0;
		_bufferedCountSum = 0;
		_unbufferedCountSum = 0;
		_writtenBytes = 0;
		_containerId2Cache.clear();
		_disabledCacheIdentifications.clear();
		_defaultBufferSize = argumentList.fetchArgument("-cachePufferGroesse=" + DEFAULT_BUFFER_SIZE).intValue();
		if(_defaultBufferSize < MINIMUM_BUFFER_SIZE) _defaultBufferSize = MINIMUM_BUFFER_SIZE;
		if(_defaultBufferSize > MAXIMUM_BUFFER_SIZE) _defaultBufferSize = MAXIMUM_BUFFER_SIZE;
		if(_debug == null) _debug = Debug.getLogger();
		String info = "Cache wurde initialisiert. Puffergröße wurde auf " + _defaultBufferSize + " Bytes je Container eingestellt.";
		if(!_cacheEnabled) info += " Cache ist momentan deaktiviert.";
		_debug.info(info);
	}

	/**
	 * Schreibt alle in den einzelnen Caches vorhandenen Daten und schließt die Cache-Verwaltung. Vor einer erneuten Verwendung muss die Cache-Verwaltung mit einem
	 * erneuten Aufruf der Methode <code>init</code> initialisiert werden.
	 */
	public void close() {
		if(DEBUG_CALLS) System.out.println("CacheManager.close");
		if(_debug == null) _debug = Debug.getLogger();
		_debug.info("Cache wird geschlossen. Alle Puffer werden auf Platte geschrieben.");

		final Collection<Cache> caches;
		_disabledCacheIdentifications.clear();
		synchronized(_containerId2Cache) {
			caches = new ArrayList<Cache>(_containerId2Cache.values());
			_containerId2Cache.clear();
		}
		int goodFlushCount = 0;
		int badFlushCount = 0;
		for(Cache cache : caches) {
			try {
				cache.flush();
				aggregateCounts(cache);
				goodFlushCount++;
			}
			catch(PersistenceException e) {
				badFlushCount++;
				_debug.warning("Fehler beim Schreiben von gecachten Daten eines Containers: " + cache, e);
			}
		}
		String message = "CacheManager wurde geschlossen.\n";
		if(badFlushCount == 0) {
			message += "Alle " + goodFlushCount + " Puffer wurden auf Platte geschrieben.\n";
		}
		else {
			message += goodFlushCount + " Puffer wurden auf Platte geschrieben. " + badFlushCount + " Puffer konnten nicht geschrieben werden\n";
		}
		message += _processedCountSum + " Datenblöcke wurden verarbeitet.\n";
		message += _bufferedCountSum + " Datenblöcke wurden zwischengespeichert.\n";
		message += _unbufferedCountSum + " Datenblöcke wurden nicht zwischengespeichert.\n";
		message += _writeCountSum + " Schreibzugriffe.";
		_debug.info(message);
	}

	/** Schreibt alle in den einzelnen Caches vorhandenen Daten. */
	public void flushAll() {
		if(DEBUG_CALLS) System.out.println("CacheManager.flushAll");
		if(_debug == null) _debug = Debug.getLogger();
		_debug.info("Alle Puffer werden auf Platte geschrieben.");

		final Collection<Cache> caches;
		synchronized(_containerId2Cache) {
			caches = new ArrayList<Cache>(_containerId2Cache.values());
		}
		int goodFlushCount = 0;
		int badFlushCount = 0;
		for(Cache cache : caches) {
			try {
				cache.flush();
				goodFlushCount++;
			}
			catch(PersistenceException e) {
				badFlushCount++;
				_debug.warning("Fehler beim Schreiben von gecachten Daten eines Containers: " + cache, e);
			}
		}
		String message = "";
		if(badFlushCount == 0) {
			message += "Alle " + goodFlushCount + " Puffer wurden auf Platte geschrieben.\n";
		}
		else {
			message += goodFlushCount + " Puffer wurden auf Platte geschrieben. " + badFlushCount + " Puffer konnten nicht geschrieben werden\n";
		}
		_debug.info(message);
	}

	/**
	 * Aktualisiert die Statistiken zu den geschriebenen Puffern
	 */
	public void updateCounts() {
		synchronized(_containerId2Cache) {
			for(Cache cache : _containerId2Cache.values()) {
				aggregateCounts(cache);
			}
		}
	}

	/**
	 * Setzt die Statistiken zu den geschriebenen Puffern auf 0 zurück
	 */
	public void resetCounts() {
		_writeCountSum = 0;
		_processedCountSum = 0;
		_bufferedCountSum = 0;
		_unbufferedCountSum = 0;
		_writtenBytes = 0;
	}

	/**
	 * Liefert den Cache für den angegebenen Container zurück.
	 *
	 * @param containerFile Container, dessen Cache bestimmt werden soll
	 *
	 * @return Cache des Containers oder <code>null</code>, wenn der Container keinen Cache hat.
	 */
	public Cache getCache(final ContainerFile containerFile) {
		if(DEBUG_CALLS) System.out.println("CacheManager.getCache " + containerFile.getContainerId());
		synchronized(_containerId2Cache) {
			return _containerId2Cache.get(containerFile.getContainerId());
		}
	}

	/**
	 * Erzeugt einen neuen Cache für den angegebenen Container.
	 *
	 * @param cont       Container für den ein neuer Cache erzeugt werden soll.
	 * @param createFile <code>true</code>, falls eine neue Datei erzeugt werden soll; <code>false</code>, falls zu einer vorhandenen Datei hinzugefügt werden
	 *                   soll.
	 *
	 * @return Neuer Cache oder <code>null</code>, falls die Datenidentifikation des Containers nicht gecacht werden soll..
	 */
	public Cache createCache(final ContainerFile cont, final boolean createFile) {
		if(!_cacheEnabled) return null;
		final IdDataIdentification identificationIds = new IdDataIdentification(cont.getObjId(), cont.getAtgId(), cont.getAspId(), cont.getSv());
		synchronized(_disabledCacheIdentifications) {
			if(_disabledCacheIdentifications.contains(identificationIds)) {
				if(DEBUG_CALLS) System.out.println("CacheManager.createCache " + identificationIds + " wird nicht gecacht");
				return null;
			}
		}
		if(DEBUG_CALLS) System.out.println("CacheManager.createCache " + identificationIds + " wird gecacht");
		final Cache cache = new Cache(cont.getContFile(), createFile, _defaultBufferSize);
		synchronized(_containerId2Cache) {
			_containerId2Cache.put(cont.getContainerId(), cache);
		}
		return cache;
	}

	/**
	 * Entfernt den Cache des angegebenen Containers aus der Containerverwaltung. Eventuell noch im Cache befindliche Daten werden nicht automatisch geschrieben,
	 * sondern müssen vorher durch einen expliziten Aufruf der Methode {@link de.bsvrz.ars.ars.persistence.CacheManager.Cache#flush()} geschrieben werden.
	 *
	 * @param containerFile Container dessen Cache aus der Verwaltung entfernt werden soll.
	 */
	public void forgetCache(final ContainerFile containerFile) {
		forgetCache(containerFile.getContainerId());
	}

	/**
	 * Entfernt den Cache des angegebenen Containers aus der Containerverwaltung. Eventuell noch im Cache befindliche Daten werden nicht automatisch geschrieben,
	 * sondern müssen vorher durch einen expliziten Aufruf der Methode {@link de.bsvrz.ars.ars.persistence.CacheManager.Cache#flush()} geschrieben werden.
	 *
	 * @param containerId ID des Containers dessen Cache aus der Verwaltung entfernt werden soll.
	 */
	public void forgetCache(final long containerId) {
		if(DEBUG_CALLS) System.out.println("CacheManager.forgetCache " + containerId);
		synchronized(_containerId2Cache) {
			final CacheManager.Cache cache = _containerId2Cache.remove(containerId);
			if(cache != null) aggregateCounts(cache);
		}
	}

	/**
	 * Überträgt, die verschiedenen Ereigniszähler eines Cache-Objekts in die entsprechenden Summenzähler der Cache-Verwaltung.
	 *
	 * @param cache Cache-Objekt, dessen Zählerstände übernommen werden sollen.
	 */
	private void aggregateCounts(final Cache cache) {
		_writeCountSum += cache._writeCount;
		_processedCountSum += cache._processedCount;
		_bufferedCountSum += cache._bufferedCount;
		_unbufferedCountSum += cache._unbufferedCount;
		_writtenBytes += cache._writtenBytes;
	}

	/**
	 * Liefert ein Array mit folgenden statistischen Informationen:
	 * <p>
	 * Gesamtanzahl von Datei-Schreibvorgängen aller geschlossenen Cache-Objekte.
	 * <p>
	 * Gesamtanzahl von verarbeiteten Datenblöcken aller geschlossenen Cache-Objekte.
	 * <p>
	 * Gesamtanzahl von zwischengespeicherten Datenblöcken aller geschlossenen Cache-Objekte.
	 * <p>
	 * Gesamtanzahl von nicht zwischengespeicherten Datenblöcken aller geschlossenen Cache-Objekte.
	 *
	 * @return Array mit statistischen Informationen
	 */
	public long[] getCounts() {
		synchronized(_containerId2Cache) {
			return new long[]{_writeCountSum, _processedCountSum, _bufferedCountSum, _unbufferedCountSum};
		}
	}

	/**
	 * Mit dieser Methode kann die Zwischenspeicherung für eine angegebene Datenidentifikation ein- bzw. ausgeschaltet werden.
	 *
	 * @param objectId          Objekt-ID der Datenidentifikation.
	 * @param atgId             ID der Attribugruppe der Datenidentifikation.
	 * @param aspectId          ID des Aspekts der Datenidentifikation.
	 * @param simulationVariant Simulationsvariante der Datenidentifikation.
	 * @param enable            <code>true</code>, wenn die Zwischenspeicherung eingeschaltetwerden soll; <code>false</code>, wenn die Zwischenspeicherung
	 *                          ausgeschaltet werden soll.
	 */
	public void setCachingEnabled(final long objectId, final long atgId, final long aspectId, final int simulationVariant, final boolean enable) {
		final IdDataIdentification identificationIds = new IdDataIdentification(objectId, atgId, aspectId, simulationVariant);
		if(DEBUG_CALLS) System.out.println("CacheManager.setCachingEnabled " + identificationIds + (enable ? " ENABLE" : " DISABLE"));
		synchronized(_disabledCacheIdentifications) {
			if(enable) {
				_disabledCacheIdentifications.remove(identificationIds);
			}
			else {
				_disabledCacheIdentifications.add(identificationIds);
			}
		}
	}


	public boolean isCacheEnabled() {
		return _cacheEnabled;
	}

	public void setCacheEnabled(final boolean cacheEnabled) {
		_cacheEnabled = cacheEnabled;
	}

	long getWriteCountSum() {
		return _writeCountSum;
	}

	long getProcessedCountSum() {
		return _processedCountSum;
	}

	long getBufferedCountSum() {
		return _bufferedCountSum;
	}

	long getUnbufferedCountSum() {
		return _unbufferedCountSum;
	}

	public long getWrittenBytes() {
		return _writtenBytes;
	}

	/** Realisiert den Zwischenspeicher für eine Containerdatei. */
	public static class Cache {

		/** Datei in die zwischengespeicherte Daten letztendlich geschrieben werden sollen. */
		private final File _file;

		/**
		 * <code>true</code>, falls beim nächsten Schreibvorgang eine neue Datei erzeugt werden soll; <code>false</code>, false beim nächsten Schreibvorgang an die
		 * bereits vorhandenen Datei angehangen werden soll.
		 */
		private boolean _createFile;

		/** Kapazität des Zwischenspeichers in Byte */
		private final int _bufferSize;

		/** Zwischenspeicher für noch zu schreibende Bytes. */
		private final ByteArrayOutputStream _out;

		/** Größe der ContainerDatei wie sie nach dem Schreiben der noch zwischengespeicherten Daten wäre. */
		private int _containerSize;

		/** Anzahl der bisher durchgeführten Datei-Schreibvorgänge in diesem Container. */
		private int _writeCount = 0;

		/** Anzahl der bisher von diesem Cache bearbeiteten Datenblöcke. */
		private int _processedCount = 0;

		/** Anzahl der bisher von diesem Cache zwischengespeicherten Datenblöcke. */
		private int _bufferedCount = 0;

		/** Anzahl der bisher von diesem Cache nicht zwischengespeicherten Datenblöcke. */
		private int _unbufferedCount = 0;

		/* Anzahl der geschriebenen Bytes */
		private long _writtenBytes = 0;


		/** @return Datei in die zwischengespeicherte Daten letztendlich geschrieben werden sollen. */
		File getFile() {
			return _file;
		}

		/**
		 * @return <code>true</code>, falls beim nächsten Schreibvorgang eine neue Datei erzeugt werden soll; <code>false</code>, false beim nächsten Schreibvorgang
		 *         an die bereits vorhandenen Datei angehangen werden soll.
		 */
		boolean isCreateFile() {
			return _createFile;
		}

		/** @return Kapazität des Zwischenspeichers in Byte */
		int getBufferSize() {
			return _bufferSize;
		}

		/** @return Anzahl der bisher durchgeführten Datei-Schreibvorgänge in diesem Container. */
		int getWriteCount() {
			return _writeCount;
		}

		/** @return Anzahl der bisher von diesem Cache bearbeiteten Datenblöcke. */
		int getProcessedCount() {
			return _processedCount;
		}

		/** @return Anzahl der bisher von diesem Cache zwischengespeicherten Datenblöcke. */
		int getBufferedCount() {
			return _bufferedCount;
		}

		/** @return Anzahl der bisher von diesem Cache nicht zwischengespeicherten Datenblöcke. */
		int getUnbufferedCount() {
			return _unbufferedCount;
		}

		/**
		 * Erzeugt einen neuen Cache für die angegebene Datei.
		 *
		 * @param file       Datei in die zwischengespeicherte Daten letztendlich geschrieben werden sollen.
		 * @param createFile <code>true</code>, falls eine neue Datei erzeugt werden soll; <code>false</code>, false an die bereits vorhandenen Datei angehangen
		 *                   werden soll.
		 * @param bufferSize Kapazität des Zwischenspeichers in Bytes.
		 */
		private Cache(final File file, final boolean createFile, int bufferSize) {
			if(DEBUG_CALLS) System.out.println("CacheManager$Cache.Cache " + file + (createFile ? " create " : " append ") + bufferSize);
			_file = file;
			_createFile = createFile;
			_bufferSize = bufferSize;
			if(_createFile) {
				_containerSize = 0;
			}
			else {
				_containerSize = (int)file.length();
			}
			_out = new ByteArrayOutputStream(bufferSize);
		}

		/**
		 * Bestimmt die Größe der ContainerDatei wie sie nach dem Schreiben der noch zwischengespeicherten Daten wäre.
		 *
		 * @return Größe der ContainerDatei in Bytes.
		 */
		public long getContainerSize() {
			if(DEBUG_CALLS) System.out.println("CacheManager$Cache.getContainerSize");
			return _containerSize;
		}

		/**
		 * Cachen der übergebenen Daten. Falls die Gesamtgröße der zwischengespeicherten Daten die Größe des Caches überschreiten würde, werden die bereits
		 * zwischengespeicherten Daten vorher in die Containerdatei geschrieben.
		 *
		 * @param data   Array mit den zu speichernden Bytes
		 * @param length Anzahl der zu speichernden Bytes im Array
		 *
		 * @throws PersistenceException Wenn beim Schreiben der Daten ein Problem aufgetreten ist.
		 */
		public void cache(byte data[], int length) throws PersistenceException {
			if(DEBUG_CALLS) System.out.println("CacheManager$Cache.cache");
			_processedCount++;
			final int cachedSize = _out.size();
			if(length + cachedSize <= _bufferSize) {
				// Block passt noch in den Cache
				_bufferedCount++;
				if(DEBUG_CALLS) System.out.println("CacheManager$Cache.cache: caching " + length);
				_out.write(data, 0, length);
			}
			else {
				// Wenn Gesamtlänge nicht in den Cache passt, wird geschrieben
				if(length >= (_bufferSize / 2)) {
					// Wenn der neue Block größer oder gleich der halben Cachegröße ist, dann wird der Cache zusammen mit dem neuen Block geschrieben
					writeCacheAndDataToFile(data, 0, length);
				}
				else {
					// Wenn der neue Block kleiner als die halbe Cachegröße ist, dann wird der Cache geschrieben und der neue Block im Cache gespeichert
					writeCacheAndDataToFile(null, 0, 0);
					_bufferedCount++;
					if(DEBUG_CALLS) System.out.println("CacheManager$Cache.cache: caching " + length);
					_out.write(data, 0, length);
				}
			}
			_containerSize += length;
		}

		/**
		 * Schreibt die zwischengespeicherten Daten in die ContainerDatei.
		 *
		 * @throws PersistenceException Wenn beim Schreiben der Daten ein Problem aufgetreten ist.
		 */
		public void flush() throws PersistenceException {
			if(DEBUG_CALLS) System.out.println("CacheManager$Cache.flush");
			if(_out.size() > 0) {
				writeCacheAndDataToFile(null, 0, 0);
			}
		}

		public String toString() {
			return "Cache für Container-Datei " + _file.toString();
		}

		/**
		 * Schreibt eventuell zwischengespeicherte Daten und optional zusätzlich einen weiteren Datensatz in die ContainerDatei
		 *
		 * @param data   Array mit weiteren zu schreibenden Daten oder <code>null</code>, falls nur die zwischengespeicherten Daten geschrieben werden sollen.
		 * @param offset Index des ersten zu schreibenden Bytes im Array
		 * @param length Anzahl der zu schreibenden Bytes im Array
		 *
		 * @throws PersistenceException Wenn beim Schreiben der Daten ein Problem aufgetreten ist.
		 */
		private void writeCacheAndDataToFile(final byte[] data, final int offset, final int length) throws PersistenceException {
			if(DEBUG_CALLS) System.out.println("CacheManager$Cache.writeCacheAndDataToFile " + _out.size() + " + " + length);
			FileOutputStream fileOutputStream = null;
			try {
				_writeCount++;
				fileOutputStream = new FileOutputStream(_file, !_createFile);
				_createFile = false;
				if(_out.size() > 0) {
					_writtenBytes += _out.size();
					// Cache schreiben
					_out.writeTo(fileOutputStream);
					_out.reset();
				}
				if(data != null) {
					_writtenBytes += length;
					_unbufferedCount++;
					fileOutputStream.write(data, offset, length);
				}
			}
			catch(FileNotFoundException e) {
				throw new PersistenceException("Container Datei zum Anhängen eines Datensatzes nicht gefunden:\n" + this);
			}
			catch(IOException e) {
				throw new PersistenceException("Archivdatensätze konnte nicht in Container-Datei geschrieben werden: " + e.getMessage() + "\n" + this);
			}
			finally {
				if(fileOutputStream != null) {
					try {
						fileOutputStream.close();
					}
					catch(IOException e) {
						throw new PersistenceException(
								"Container-Datei konnte beim Anhängen von Datensätzen nicht geschlossen werden: " + e.getMessage() + "\n" + this
						);
					}
				}
			}
		}
	}

}
