/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.persistence.index.ATimeMonotoneIndex;
import de.bsvrz.ars.ars.persistence.index.DIdxATimeMonotoneIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndex;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndex;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Diese Klasse teilt das Persistenzverzeichnis in disjunkte Teilbaeume auf und stellt alle Verwaltungsinformationen wieder her.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 * @see PersistenceManager#REBUILDINDEX_FILE_FLAG_NAME
 * @see PersistenceManager#rebuildIndex(Task, byte[], long, long, long, int, ArchiveDataKind)
 */
public class RestorePersDirTsk extends Task {

	private static int RESTORE_TEMP_MEM = 16 * 1024;

	private static RestorePersDirTsk[] restoreTasks;

	private static long startTime;

	// damit man nach dem cleanup noch zu testzwecken auf die ergebnisse zugreifen kann:
	public static int test_totalDID = 0, test_badDID = 0, test_goodDID = 0, test_viscont = 0, test_frebuilt = 0;

	private PersistenceManager persMgr;

	private List<File> startDirs;

	private byte[] tmpMem;

	private long maxArchiveTime = 0, maxContID = 0;

	private int maxMedienID = 0;

	private int totalDID = 0, badDID = 0, goodDID = 0;		// Zähler für DID

	/** Zähler für unerwartete RuntimeExceptions */
	private int _errorCount = 0;

	private int[] total_cont_rebuilt = new int[2];		// Zähler für alle besuchten Container


	/**
	 * Zerlegt das Persistenzverzeichnis in eine Anzahl disjunkter Teile, erzeugt ebenso viele Wiederherstellungstasks und startet sie. Zum Wiederherstellen
	 * gehoert: <UL> <LI>Suche nach der groessten, zuletzt verwendeten ContainerID.</LI> <LI>Suche nach der groessten, zuletzt verwendeten Archivzeit.</LI>
	 * <LI>Neuaufbau der Indexe, falls im Verzeichnis einer Datenidentifikation und Datensatzart eine {@link PersistenceManager#REBUILDINDEX_FILE_FLAG_NAME}-Datei
	 * (siehe Beschreibung dort) vorhanden ist.</LI> </UL>
	 *
	 * @param pMgr            Aufrufender PersistenceManager
	 * @param numOfPartitions gewuenschte maximale Anzahl disjunkter Teile
	 */
	public static void startRestore(PersistenceManager pMgr, int numOfPartitions) {
		startTime = System.currentTimeMillis();
		List<List<File>> partitions = RestorePersDirTsk.calcPartitions(pMgr.getArchivePath(), numOfPartitions);
		restoreTasks = new RestorePersDirTsk[partitions.size()];
		for(int i = 0; i < partitions.size(); i++) {
			restoreTasks[i] = new RestorePersDirTsk(pMgr, partitions.get(i), RESTORE_TEMP_MEM);	// startet auch gleich
		}
	}

	/** @return Wahr, falls alle Tasks fertig sind, falsch sonst */
	public static boolean restoreDone() {
		for(RestorePersDirTsk rt : restoreTasks) {
			if(!rt.isTerminated()) return false;
		}
		return true;
	}

	/** Bricht den Wiederherstellungslauf ab */
	public static void terminateRestore() {
		for(int i = 0; i < restoreTasks.length; i++) {
			restoreTasks[i].terminateTask();
		}
	}

	/** Gibt den benutzten Speicher frei (insbes. den von den einzelnen Tasks zum Wiederherstellen der Indexe benoetigten Speicher) */
	public static void cleanUpTask() {
		

		// damit man nach dem cleanup noch zu testzwecken auf die ergebnisse zugreifen kann:
		test_totalDID = getTotalDID();
		test_badDID = getBadDID();
		test_goodDID = getGoodDID();
		test_viscont = getVisitedCont();
		test_frebuilt = getFullyRebuiltCont();

		for(RestorePersDirTsk rt : restoreTasks) {
			rt.persMgr.removeContainerFileObjects(rt);
		}
		restoreTasks = null;
	}

	/** @return Maximum der von allen Threads gefundenen Archivzeit */
	public static long getMaxArchiveTime() {
		long maxAT = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			maxAT = Math.max(maxAT, rt.maxArchiveTime);
		}
		return maxAT;
	}

	/** @return Maximum der von allen Threads gefundenen Container-ID */
	public static long getMaxContID() {
		long maxCID = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			maxCID = Math.max(maxCID, rt.maxContID);
		}
		return maxCID;
	}

	/** @return Maximum der von allen Threads gefundenen Medien-ID */
	public static int getMaxMedienID() {
		int maxMID = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			maxMID = Math.max(maxMID, rt.maxMedienID);
		}
		return maxMID;
	}

	/** @return Summe der von allen Threads besuchten DID-Verzeichnisse */
	public static int getTotalDID() {
		int total = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			total += rt.totalDID;
		}
		return total;
	}

	/** @return Summe der von allen Threads besuchten DID-Verzeichnisse, in denen die Wiederherstellung erfolgreich abgeschlossen wurde. */
	public static int getGoodDID() {
		int good = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			good += rt.goodDID;
		}
		return good;
	}

	/** @return Summe der von allen Threads besuchten DID-Verzeichnisse, in denen die Wiederherstellung nicht erfolgreich abgeschlossen wurde. */
	public static int getBadDID() {
		int bad = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			bad += rt.badDID;
		}
		return bad;
	}

	/** @return Summe der in allen Threads aufgetretenen RuntimeExceptions während der Wiederherstellung. */
	public static int getErrorCount() {
		int errorCount = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			errorCount += rt._errorCount;
		}
		return errorCount;
	}

	/** @return Zahl der bisher besuchten Container */
	public static int getVisitedCont() {
		int vc = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			vc += rt.total_cont_rebuilt[0];
		}
		return vc;
	}

	/** @return Zahl der bisher besuchten Container, bei denen alle Datensätze zur Wiederherstellung gelesen werden mussten */
	public static int getFullyRebuiltCont() {
		int frc = 0;
		for(RestorePersDirTsk rt : restoreTasks) {
			frc += rt.total_cont_rebuilt[1];
		}
		return frc;
	}

	/** @return Wahr, falls kein Verzeichnis nicht wiederhergestellt werden konnte, falsch sonst */
	public static boolean success() {
		return getBadDID() == 0 && getErrorCount() == 0;
	}

	/** @return Zusammenfassung des Wiederherstellungslaufes */
	public static String summary() {
		int good = getGoodDID(), bad = getBadDID();
		int errorCount = getErrorCount();
		return "Wiederherstellungslauf beendet.\n  " + getTotalDID() + " Container-Verzeichnisse besucht, "
		       + (good + bad != 0 ? ("bei " + (good + bad) + " Index-Neuaufbau notwendig, davon " + good + " erfolgreich und " + bad + " nicht erfolgreich."
		                             + "\n  " + getVisitedCont() + " Container erfolgreich besucht, bei " + getFullyRebuiltCont()
		                             + " davon vollst. Durchlauf notwendig") : "kein Index-Neuaufbau notwendig")
		       + (errorCount != 0 ? ("\nAufgetretene RuntimeExceptions: " + errorCount) : "")
		       + "\nZeit: " + Util.relTimestrMillis( System.currentTimeMillis() - startTime
		);
	}

	/** @return Zwischenstand des Wiederherstellungslaufes */
	public static String status() {
		int good = getGoodDID(), bad = getBadDID();
		int errorCount = getErrorCount();
		return "Wiederherstellung laeuft...\n" + getTotalDID() + " Container-Verzeichnisse besucht, " + "bei " + (good + bad)
		       + " Index-Neuaufbau notwendig, davon " + good + " erfolgreich und " + bad + " nicht erfolgreich durchgefuehrt." + "\n  " + getVisitedCont()
		       + " Container erfolgreich besucht, bei " + getFullyRebuiltCont() + " davon vollst. Durchlauf notwendig"
		       + (errorCount != 0 ? ("\nAufgetretene RuntimeExceptions: " + errorCount) : "")
		       + "\nZeit bisher: " + Util.relTimestrMillis(System.currentTimeMillis() - startTime);
	}

	private RestorePersDirTsk(PersistenceManager pMgr, List<File> startDirectories, int tmpMemSize) {
		super(null);
		persMgr = pMgr;
		startDirs = startDirectories;
		tmpMem = new byte[tmpMemSize];
		start();
	}

	/** Durchlaueft mit Hilfe eines {@link ContainerDataIterator} jedes Verzeichnis aus {@link #startDirs} und stellt die Verwaltungsinformation wieder her. */
	@Override
	public void execute() {
		try {
			for(File dir : startDirs) {
				File didDir;
				ContainerFileDirIterator iter = new ContainerFileDirIterator(dir.getAbsolutePath());
				while((didDir = iter.getNextDIdDirectory()) != null) {
					try {
						if(shouldTerminate()) return;
						totalDID++;

						// Indexe ggf. wiederherstellen:
						File changedFlagFile = new File(didDir, PersistenceManager.REBUILDINDEX_FILE_FLAG_NAME);
						boolean changedFlagFileExists = changedFlagFile.exists();

						// Strategie: Wenn die _rebuildIndex.flag-Datei vorhanden ist oder einer der Indexe
						// fehlt, werden die Indexe neu aufgebaut.

						if(changedFlagFileExists || anyIndexFileMissing(didDir)) {
							try {
								long t2 = Util.startTimer();
								int[] cont_rebuilt = persMgr.rebuildIndex(this, tmpMem, didDir.getAbsolutePath());
								total_cont_rebuilt[0] += cont_rebuilt[0];
								total_cont_rebuilt[1] += cont_rebuilt[1];
								if(changedFlagFileExists && !changedFlagFile.delete()) {
									badDID++;
									_debug.error(
											"Wiederherstellungslauf: '" + PersistenceManager.REBUILDINDEX_FILE_FLAG_NAME + "' im Verzeichnis '" + didDir
											+ "' konnte nicht geloescht werden."
									);
								}
								else {
									goodDID++;
									_debug.fine("Wiederherstellungslauf: Indexe in " + didDir + " erfolgreich neu aufgebaut (" + Util.stopTimer(t2) + ").");
								}
							}
							catch(Exception e) {
								badDID++;
								_debug.error("Wiederherstellungslauf: Index der Datenidentifikation '" + didDir + "' konnte nicht neu angelegt werden", e);
							}
						}

						// maximale Archivzeit. Medien- und Container-ID ermitteln
						long[] sortedContainerIDs = PersistenceManager.getSortedContainerIDs(didDir.getAbsolutePath());
						long largestContID = sortedContainerIDs.length == 0 ? -1 : sortedContainerIDs[sortedContainerIDs.length - 1];
						if(largestContID > 0) {
							maxContID = Math.max(maxContID, largestContID);
							long[] aTimeMedID = getMaxArchiveTimeMedID(didDir, sortedContainerIDs);
							maxArchiveTime = Math.max(maxArchiveTime, aTimeMedID[0]);
							maxMedienID = Math.max(maxMedienID, (int)aTimeMedID[1]);
						}
					}
					catch(RuntimeException e) {
						_errorCount++;
						_debug.error("Wiederherstellungslauf: Unbehandelter Fehler in RestorePersDirTsk im Verzeichnis " + didDir, e);
					}
				}
			}
		}
		catch(RuntimeException e) {
			_errorCount++;
			_debug.error("Wiederherstellungslauf: Unbehandelter Fehler in RestorePersDirTsk", e);
		}
		finally {
			super.execute();
		}
	}

	/**
	 * Prueft, ob im genannten Verzeichnis einer DId und Datenart mindestens eine Indexdatei fehlt.
	 *
	 * @param didDir Verzeichnis das geprüft werden soll
	 *
	 * @return Wahr wenn ein Index fehlt, falsch sonst
	 */
	private boolean anyIndexFileMissing(File didDir) {
		boolean requested = PersistenceManager.getArchiveDataKindFromDir(didDir).isRequested();
		if(!new File(didDir, DataTimeIndex.IDX_FILENAME).exists()) return true;
		if(requested) {
			if(!new File(didDir, ATimeMonotoneIndex.IDX_FILENAME).exists()) return true;
			if(!new File(didDir, DataIndexIndex.IDX_FILENAME).exists()) return true;
		}
		else if(!new File(didDir, DIdxATimeMonotoneIndex.IDX_FILENAME).exists()) return true;
		return false;
	}

	/**
	 * Gibt die maximale Archivzeit und MedienID aus einem Verzeichnis mit Containerdateien zurück. Die Archivzeit wird aus dem Header oder den Archivdaten der
	 * jüngsten Containerdatei gelesen. Da der letzte Container häufig nicht abgeschlossen ist, werden zur Ermittlung der maximalen MedienID auch die Header der
	 * älteren Containerdateien geprüft.
	 *
	 * @param didDir       Verzeichnis mit den Containerdateien
	 * @param containerIds Aufsteigend sortiertes Array mit den ContainerIDs der in diesem Verzeichnis enthaltenen Container (kann über {@link
	 *                     PersistenceManager#getSortedContainerIDs(String)} ermittelt werden}
	 *
	 * @return Long-Array mit 2 Einträgen, wobei der erste die Archivzeit und der zweite die MedienID darstellt. {-1, -1} wird zurückgeliefert, wenn ein leeres
	 *         Array übergeben wurde oder ein Fehler auftrat. Wenn nur die MedienID nicht ermittelt werden konnte, enthält das Array entsprechend {Archivzeit,
	 *         -1}.
	 */
	private long[] getMaxArchiveTimeMedID(File didDir, long[] containerIds) {
		try {
			String medienID;
			long aTime;
			final int containerIdsLength = containerIds.length;
			if(containerIdsLength == 0) return new long[]{-1, -1};

			// Zuerst den jüngsten Container nach ArchiveTime und MedienID durchsuchen

			long contId = containerIds[containerIdsLength - 1];

			persMgr.accessContainerReadOnly(this, new File(didDir, ContainerFile.getContainerFileName(contId)));
			try {
				persMgr.readContainerHeader(this);
				aTime = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_ARC_TIME_MAX);
				medienID = persMgr.getContainerHeaderParamAsString(this, ContainerHdr.CHP_MEDIUM_ID);

				// der letzte Container ist womöglich nicht abgeschlossen, darum Datensätze durchsuchen:
				if(aTime <= 0) {
					//Da aTime <= 0 ist, ist der Container nicht abgeschlossen, daher muss nicht auf evtl. gelöschte Container Rücksicht genommen werden
					ContainerDataIterator cdi = persMgr.getArchiveData(this);
					while(cdi.readNextDataHdr()) {
						aTime = Math.max(aTime, cdi.getArcTime());
						cdi.skipData();
					}
					cdi.close();
				}
			}
			finally {
				persMgr.leaveContainer(this);
			}
			// Falls der letzte Container bereits eine gültige MedienID hat, kann hier aufgehört werden.
			if(!medienID.equals(ContainerHdr.MEDIUM_UNSAVED)) {
				return new long[]{aTime, Integer.parseInt(medienID)};
			}


			// Versuchen die maximale MedienID aus der ContainerBackupIndex-Datei auszulesen
			ContainerBackupIndex containerBackupIndex = ContainerBackupIndex.getContainerBackupIndex(didDir);
			try {
				try {
					if(containerBackupIndex != null) {
						final int id = containerBackupIndex.getMaxMedienID();
						if(id == ContainerBackupIndex.MEDIUM_ID_UNSAVED) {
							// Datenidentifikation wurde noch nie gesichert.
							return new long[]{aTime, -1};
						}
						else if(id != ContainerBackupIndex.MEDIUM_ID_UNKNOWN) {
							// Datenidentifikation wurde schon gesichert.
							return new long[]{aTime, id};
						}
					}
				}
				catch(IOException e) {
					_debug.warning("Konnte nicht auf ContainerBackupIndex zugreifen", e);
				}

				if(containerBackupIndex == null) containerBackupIndex = createContainerBackupIndex(didDir);

				// Auslesen aus der ContainerBackupIndex ist fehlgeschlagen,
				// nun auch die älteren Containerdateien der Reihe nach durchsuchen
				for(int i = containerIdsLength - 2; i >= 0; i--) {
					contId = containerIds[i];
					persMgr.accessContainerReadOnly(this, new File(didDir, ContainerFile.getContainerFileName(contId)));
					try {
						persMgr.readContainerHeader(this);
						medienID = persMgr.getContainerHeaderParamAsString(this, ContainerHdr.CHP_MEDIUM_ID);
					}
					finally {
						persMgr.leaveContainer(this);
					}
					if(!medienID.equals(ContainerHdr.MEDIUM_UNSAVED)) {
						// gefundene MedienID in den ContainerBackupIndex eintragen
						final int foundID = Integer.parseInt(medienID);
						try {
							if(containerBackupIndex != null) containerBackupIndex.setMaxMedienID(foundID);
						}
						catch(IOException e) {
							_debug.warning("Konnte MedienID nicht in ContainerBackupIndex eintragen: " + didDir, e);
							if(!ContainerBackupIndex.deleteContainerBackupIndex(didDir)) {
								_debug.error("Konnte fehlerhafte ContainerBackupIndex-Datei nicht löschen: " + didDir);
							}
						}
						return new long[]{aTime, foundID};
					}
				}
				try {
					if(containerBackupIndex != null) containerBackupIndex.setMaxMedienID(ContainerBackupIndex.MEDIUM_ID_UNSAVED);
				}
				catch(IOException e) {
					_debug.warning("Konnte nicht in ContainerBackupIndex eintragen, dass Datenidentifikation noch nie gesichert wurde: " + didDir, e);
				}
			}
			finally {
				if(containerBackupIndex != null) containerBackupIndex.close();
			}
			return new long[]{aTime, -1};
		}
		catch(PersistenceException e) {
			_debug.error("Wiederherstellungslauf: Fehler beim Rekonstruieren von letzter Archivzeit/MedienID: " + e.getMessage());
			return new long[]{-1, -1};
		}
	}

	/**
	 * Erstellt einen ContainerBackupIndex mit den zu sichernden ContainerIDs in einem Verzeichnis
	 *
	 * @param didDir Verzeichnis
	 *
	 * @return ContainerBackupIndex
	 */
	private ContainerBackupIndex createContainerBackupIndex(final File didDir) {
		try {
			final List<Long> containerIDs = BackupTask.getContainerIDsToBackupStatic(didDir, persMgr, this);

			final ContainerBackupIndex newContainerBackupIndex = ContainerBackupIndex.createNewContainerBackupIndex(didDir);

			for(Long containerID : containerIDs) {
				newContainerBackupIndex.appendContainerID(containerID);
			}
			return newContainerBackupIndex;
		}
		catch(PersistenceException e) {
			_debug.warning("Fehler beim Anlegen eines ContainerBackupIndex:" + didDir, e);
		}
		catch(IOException e) {
			_debug.warning("Fehler beim Anlegen eines ContainerBackupIndex:" + didDir, e);
		}
		return null;
	}

	/**
	 * Teilt das uebergebene Verzeichnis in eine Liste (mit <code>numOfPartitions</code> Elementen) von Listen von Verzeichnissen ein. Diese Liste repraesentiert
	 * das vollstaendige Verzeichnis. Jedes Element kann von einem eigenen Thread bearbeitet werden.
	 *
	 * @param persDir
	 * @param numOfPartitions
	 *
	 * @return
	 */
	@SuppressWarnings({"unchecked"})
	private static List<List<File>> calcPartitions(String persDir, int numOfPartitions) {
		List<List<File>> dirs = Util.mkList(Util.mkList(new File(persDir)));
		while(true) {
			boolean splitted = false;
			int k = 0;
			while(k < dirs.size() && dirs.size() < numOfPartitions) {
				stepDown(dirs.get(k));
				if(splitDirs(dirs, k)) {
					k++;
					splitted = true;
				}
				k++;
			}
			if(!splitted) return dirs;
		}
	}

	/**
	 * Steigt solange die Verzeichnishierarchie hinab (maximal bis zur sv-Ebene), wie stets nur ein Verzeichnis auf dem Pfad liegt. Das unterste einzelne
	 * Verzeichnis wird dann durch die enthaltenen Unterverzeichnisse ersetzt.
	 *
	 * @param dirs Verzeichnisse eines Verzeichnisses
	 */
	private static void stepDown(List<File> dirs) {
		while(dirs.size() == 1 && !PersistenceManager.isValidSvFolderName(dirs.get(0).getName())) {
			File[] subdirs = Util.listDirectories(dirs.get(0));
			dirs.clear();
			for(File file : subdirs) {
				dirs.add(file);
			}
		}
	}

	/**
	 * Die uebergebene Liste enthaelt einen vollstaendigen Querschnitt durch die Verzeichnishierarchie. Diese Methode teilt die Liste von Verzeichnissen an
	 * Position <code>pos</code> in zwei Haelften und fuegt die zweite Haelfte hinter <code>pos</code> ein.
	 *
	 * @param dirs Liste von Listen von Verzeichnissen
	 * @param pos  Position
	 *
	 * @return Wahr, falls die Verzeichnisliste geteilt wurde.
	 */
	private static boolean splitDirs(List<List<File>> dirs, int pos) {
		List<File> subdirs = dirs.get(pos);
		if(subdirs.size() >= 2) {
			dirs.add(pos + 1, new ArrayList<File>(subdirs.subList(subdirs.size() / 2, subdirs.size())));
			subdirs.removeAll(dirs.get(pos + 1));
			return true;
		}
		return false;
	}
}
