/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;


/**
 * Klasse zum Lesen und Schreiben der Startup-Info, die das Archivsystem braucht, um starten zu können. Wenn die Startup-Info nicht vorliegt, muss sie neu
 * erzeugt werden (aufwendig).
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class StartupProperties extends KeyValProps {

	public static String STARTUP_PROP_FILE_NAME = "_startUpInfo.property";

	/** Sortierte Liste aller StartupInfo-Parameter. */
	protected static final List<KeyValParam> allParams = new ArrayList<KeyValParam>();

	private File startUpPropFile;

	private static final long INVALID_CONTAINER_ID = -1;

	private static final long INVALID_TIME = -1;

	private static final long INVALID_MEDIENID = -1;

	public static final KeyValParam STUP_MAX_CONT_ID = createNumParam(
			"maxContainerID", INVALID_CONTAINER_ID, BYTES5_STR_LEN, INVALID_CONTAINER_ID, BYTE5_MAXVAL
	), STUP_LAST_ATIME = createNumParam("lastArchiveTime", INVALID_TIME, BYTES6_STR_LEN, INVALID_TIME, BYTE6_MAXVAL), STUP_LAST_USED_MEDIUM_ID = createNumParam(
			"lastUsedMediumID", INVALID_MEDIENID, BYTES4_STR_LEN, INVALID_MEDIENID, BYTE4_MAXVAL
	);

	/** Laenge des Key/Value-Blocks in Byte. */
	public final static int HDR_TXT_LEN;

	static {
		allParams.add(STUP_MAX_CONT_ID);
		allParams.add(STUP_LAST_ATIME);
		allParams.add(STUP_LAST_USED_MEDIUM_ID);

		int headerLen = 0;
		for(KeyValParam param : allParams) {
			headerLen += param.getKey().length() + "=".length() + param.getValLen() + CH_DELIM.length();
		}
		HDR_TXT_LEN = headerLen;
	}


	/**
	 * Initialisiert die StartUp-Properties.
	 *
	 * @param persBaseDir Wurzel des Persistenzverzeichnisses
	 */
	public StartupProperties(String persBaseDir) {
		super();
		startUpPropFile = new File(persBaseDir, STARTUP_PROP_FILE_NAME);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer startup.property-Datei ist
	 */
	public static boolean isStartupInfoFile(String name) {
		return name.equals(STARTUP_PROP_FILE_NAME);
	}

	/**
	 * Liest die StartUp-Properties ein
	 *
	 * @throws PersistenceException, falls ein Fehler aufgetreten ist. In diesem Fall muss der Restore-Vorgang gestartet werden.
	 */
	public void readStartUpProperties() throws PersistenceException {
		properties.clear();
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(startUpPropFile);
			Properties tmpProp = new Properties();	// zur Pruefung erst in extra Properties laden
			tmpProp.load(fis);
			for(KeyValParam param : allParams) {
				setVal(param, (String)tmpProp.get(param.getKey()));
			}
		}
		catch(IOException e) {
			throw new PersistenceException("Startup-Properties konnten nicht gelesen werden: " + e.getMessage());
		}
		finally {
			try {
				if(fis != null) fis.close();
			}
			catch(IOException e) {
				throw new PersistenceException("Fehler beim Schliessen der Startup-Properties: " + e.getMessage());
			}
		}
	}

	public void writeStartUpProperties() throws PersistenceException {
		try {
			Util.deleteCreateNewFile(startUpPropFile);
			FileWriter fw = new FileWriter(startUpPropFile);
			StringBuilder sb = new StringBuilder(HDR_TXT_LEN);
			for(KeyValParam param : allParams) {
				sb.append(param.getKey() + "=" + properties.getProperty(param.getKey()) + CH_DELIM);
			}
			fw.write(sb.toString());
			fw.close();
		}
		catch(IOException e) {
			throw new PersistenceException("Startup-Properties konnten nicht geschrieben werden: " + e.getMessage());
		}
	}

	public boolean deleteStartupPropsFile() {
		return Util.deleteIfExists(startUpPropFile);
	}

	public String toString() {
		return "Startup Properties Datei '" + startUpPropFile.getAbsolutePath() + "'";
	}
}
