/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.index;


import de.bsvrz.ars.ars.persistence.index.result.ArchiveTimeIndexEntry;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;

/**
 * In dieser Klasse ist zusaetzlich gefordert, dass die Folge der Datenindexe streng monoton ist.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DIdxATimeMonotoneIndex extends ATimeMonotoneIndex {

	/**
	 * Erzeugt einen neuen monoton steigenden Index. Neue Eintraege werden an eine bestehende Indexdatei angefuegt, wenn diese bereits existiert. Existiert keine
	 * Indexdatei, wird eine erzeugt. Initial werden keine Daten eingelesen; es findet kein Plattenzugriff statt.
	 *
	 * @param size Größe des Buffers (Anzahl der Einträge). Wird mindestens auf 1 gesetzt.
	 */
	public DIdxATimeMonotoneIndex(int size) {
		super(size);
	}

	/**
	 * Erzeugt einen neuen monoton steigenden Index.
	 * <p>
	 * Bei startmode==APPEND_REUSE_LAST wird der Ringpuffer mit den aktuellsten Daten aus einer vorhandenen Indexdate gefüllt. Zusätzlich wird der letzte Eintrag
	 * in der Indexdatei zum currentEntry. Damit kann ein Container fortgesetzt und der Key2 weiter aktualisiert werden.
	 * <p>
	 * Bei startmode==CREATE_NEW wird ein leerer Index erzeugt und die bisherige Indexdatei (sofern vorhanden) geloescht.
	 *
	 * @param size      Größe des Buffers (Anzahl der Einträge)
	 * @param basePath  Verzeichnis der Indexdatei
	 * @param startmode {@link IndexStartmode#ERASE_OLD}: alte Indexdatei loeschen, {@link IndexStartmode#APPEND_REUSE_LAST}: anfuegen, letzten Eintrag
	 *                  weiterverwenden
	 *
	 * @throws IndexException
	 */
	public DIdxATimeMonotoneIndex(int size, String basePath, IndexStartmode startmode) throws IndexException {
		super(size, basePath, startmode);
	}

	/**
	 * Fügt einen neuen Eintrag in den Buffer ein. Bei Bedarf wird der Puffer auf Platte geschrieben.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 * @param dIMin    Min-Wert des Datenindex
	 * @param dIMax    Max-Wert des Datenindex
	 * @param aTMin    Min-Wert der Archivzeit
	 * @param aTMax    Max-Wert der Archivzeit
	 * @param dTMin    Min-Wert der Datenzeit
	 * @param dTMax    Max-Wert der Datenzeit
	 * @param cID      Container-ID des Neueintrags
	 *
	 * @throws IndexException Falls die Konsistenzpruefung der Eintraege fehlschlaegt
	 */
	public void add(String basePath, long dIMin, long dIMax, long aTMin, long aTMax, long dTMin, long dTMax, long cID) throws IndexException {
		checkForAddLocal(dIMin, dIMax, aTMin, aTMax, dTMin, dTMax, cID);
		super.add(basePath, dIMin, dIMax, aTMin, aTMax, dTMin, dTMax, cID);
	}

	/**
	 * Prüft, ob dem Index ein neuer Eintrag hinzugefügt werden könnte.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 * @param dIMin    Min-Wert des Datenindex
	 * @param dIMax    Max-Wert des Datenindex
	 * @param aTMin    Min-Wert der Archivzeit
	 * @param aTMax    Max-Wert der Archivzeit
	 * @param dTMin    Min-Wert der Datenzeit
	 * @param dTMax    Max-Wert der Datenzeit
	 * @param cID      Container-ID des Neueintrags
	 *
	 * @throws IndexException Falls die Konsistenzpruefung der Eintraege fehlschlaegt
	 */
	public void checkForAdd(String basePath, long dIMin, long dIMax, long aTMin, long aTMax, long dTMin, long dTMax, long cID) throws IndexException {
		checkForAddLocal(dIMin, dIMax, aTMin, aTMax, dTMin, dTMax, cID);
		super.checkForAdd(basePath, dIMin, dIMax, aTMin, aTMax, dTMin, dTMax, cID);
	}

	private void checkForAddLocal(final long dIMin, final long dIMax, final long aTMin, final long aTMax, final long dTMin, final long dTMax, final long cID)
			throws CorruptIndexException {
		// Neue Werte pruefen:
		try {
			if(getCurrentDIMax() >= dIMin) throw new Exception("Neueintrag nicht streng monoton: currentDIdxMax=" + getCurrentDIMax());
		}
		catch(Exception e) {
			throw new CorruptIndexException(
					e.getMessage() + " (" + "dIMin=" + dIMin + ", dIMax=" + dIMax + ", aTMin=" + aTMin + ", aTMax=" + aTMax + ", dTMin=" + dTMin + ", dTMax="
					+ dTMax + ", contID=" + cID + ")",
					e
			);
		}
	}


	public IndexResult<ArchiveTimeIndexEntry> getContainerIDByDIdx(String basePath, byte[] temp, long minKey, long maxKey) throws IndexException {
		return getContainerID(basePath, temp, DI, minKey, maxKey);
	}

	/**
	 * Liefert den kleinsten Datenindex-Minwert, der jemals in diesem Index gespeichert wurde.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 *
	 * @return Kleinster Datenindex
	 *
	 * @throws IndexException Bei Zugriffsproblemen mit der IndexDatei
	 */
	public long getAbsoluteDIdxMin(String basePath) throws IndexException {
		IndexResult<de.bsvrz.ars.ars.persistence.index.result.ArchiveTimeIndexEntry> ir = getMinEntry(basePath);
		if(ir != null && ir.size() > 0) {
			return ir.get(0).getDataIndexMinimum();
		}
		else {
			return IndexResult.NO_VALUE;
		}
	}
}
