/*
 *
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.iter;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.persistence.*;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.List;

/**
 * TBD Dokumentation
 *
 * @author Kappich Systemberatung
 */
public class SequentialContainerIterator implements DataIterator {

	private final List<QueryDataSequence.IndexedContainer> _containerFiles;
	private final ArchiveManager _archiveManager;
	private DataIterator _dataIterator;
	private IdDataIdentification _did;
	private final DeletedTreatment _deletedTreatment;
	private Task _task;
	private int _index = 0;
	private ArchiveDataKind _adk;

	public SequentialContainerIterator(final ArchiveManager archiveManager, final List<QueryDataSequence.IndexedContainer> containerFiles, final IdDataIdentification did, final ArchiveDataKind adk, final DeletedTreatment deletedTreatment, final Task task) throws PersistenceException {
		_archiveManager = archiveManager;
		_did = did;
		_adk = adk;
		_deletedTreatment = deletedTreatment;
		_task = task;
		if(containerFiles.isEmpty()) throw new IllegalArgumentException("containerFiles");
		_containerFiles = containerFiles;
		openContainer(_index);
	}

	/** Eindeutiger ContainerFile-Index für Persistenz-Zugriff. */
	private static long getContainerFileIdx(ArchiveDataKind dataKind, long containerId) {
		int dataKindOffset = dataKind.getCode();
		long containerOffset = containerId * Util.getAllDataKinds().length;
		return dataKindOffset + containerOffset;
	}

	private void openContainer(final int index) throws PersistenceException {
		QueryDataSequence.IndexedContainer container = _containerFiles.get(index);
		final long contIndex = getContainerFileIdx(_adk, container.getContainerId());
		PersistenceManager persMgr = _archiveManager.getPersMgr();
		persMgr.accessContainer(_task, contIndex, _did, _adk, container.getContainerId());
		persMgr.readContainerHeader(_task, contIndex);
		final boolean deleted = persMgr.getContainerHeaderParamAsBoolean(_task, contIndex, ContainerHdr.CHP_DELETED);
		if(deleted) {
			if(_deletedTreatment == DeletedTreatment.Ignore) {
				// Container ignorieren
				_dataIterator = DataIterator.empty();
				return;
			}
			else if(_deletedTreatment == DeletedTreatment.MinMax) {
				// Datensätze ignorieren, nur über Header iterieren
				_dataIterator = new DataIteratorAdapter(persMgr.getHdrData(_task, contIndex), contIndex, container.getContainerId());
				return;
			}
		}
		_dataIterator = new DataIteratorAdapter(persMgr.getArchiveData(_task, contIndex, _deletedTreatment == DeletedTreatment.Full), contIndex, container.getContainerId());
	}

	private void closeContainer(final int index) throws PersistenceException {
		QueryDataSequence.IndexedContainer container = _containerFiles.get(index);
		long cfIdx = getContainerFileIdx(_adk, container.getContainerId());
		_archiveManager.getPersMgr().leaveContainer(_task, cfIdx);
	}

	@Override
	public void peek(final ContainerDataResult result) throws PersistenceException {
		_dataIterator.peek(result);
	}

	@Override
	public long peekDataIndex() {
		return _dataIterator.peekDataIndex();
	}

	@Override
	public long peekDataTime() {
		return _dataIterator.peekDataTime();
	}

	@Override
	public long peekArchiveTime() {
		return _dataIterator.peekArchiveTime();
	}

	@Override
	public void remove() throws PersistenceException {
		_dataIterator.remove();
		while(_dataIterator.isEmpty() && _index < _containerFiles.size() - 1) {
			closeContainer(_index);
			_index++;
			openContainer(_index);
		}
	}

	@Override
	public boolean isEmpty() {
		return _dataIterator.isEmpty();
	}

	@Override
	public void close() throws PersistenceException {
		_dataIterator.close();
		closeContainer(_index);
	}

	@Override
	public long getCurContFileIdx() {
		return _dataIterator.getCurContFileIdx();
	}
}
