/*
 *
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.iter;

import de.bsvrz.ars.ars.persistence.ContainerDataResult;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;

import java.util.NoSuchElementException;

/**
 * TBD Dokumentation
 *
 * @author Kappich Systemberatung
 */
public class TimeSpecificationFilterIterator implements DataIterator{

	private final DataIterator _delegate;
	private final ArchiveTimeSpecification _timeSpecification;
	private ContainerDataResult _firstData;
	private long _firstIdx;

	public TimeSpecificationFilterIterator(DataIterator delegate, ArchiveTimeSpecification timeSpecification) throws PersistenceException {
		_delegate = delegate;
		_timeSpecification = timeSpecification;
		if(_delegate.isEmpty()) {
			// Leere Anfrage
			_firstData = null;
			return;
		}
		if(!isInRange()) {
			_firstData = new ContainerDataResult();
			while(true) {
				_delegate.peek(_firstData);
				_firstIdx = _delegate.getCurContFileIdx();
				_delegate.remove();
				if(_delegate.isEmpty()) {
					return;
				}
				if(isInRange()) {
					if(isAtStart()) {
						// Wir brauchen den vorigen Datensatz nicht, weil wir uns direkt auf dem Intervallanfang befinden
						_firstData = null;
					}
					return;
				}
			}
		}
	}

	private void removeIfInvalid() throws PersistenceException {
		if(isEmpty()) return;
		while(!isInRange()) {
			_delegate.remove();
			if(isEmpty()) return;
		}
	}

	private boolean isInRange() {
		if(_timeSpecification.getTimingType() == TimingType.DATA_INDEX) {
			long dataIndex = _delegate.peekDataIndex();
			return dataIndex >= _timeSpecification.getIntervalStart() && dataIndex <= _timeSpecification.getIntervalEnd();
		}
		if(_timeSpecification.getTimingType() == TimingType.DATA_TIME) {
			long dataTime = _delegate.peekDataTime();
			if(dataTime == 0) return true;
			return dataTime >= _timeSpecification.getIntervalStart() && dataTime <= _timeSpecification.getIntervalEnd();
		}
		if(_timeSpecification.getTimingType() == TimingType.ARCHIVE_TIME) {
			long archiveTime = _delegate.peekArchiveTime();
			if(archiveTime == 0) return true;
			return archiveTime >= _timeSpecification.getIntervalStart() && archiveTime <= _timeSpecification.getIntervalEnd();
		}
		throw new AssertionError("Unbekannter TimingType: " + _timeSpecification.getTimingType());
	}

	private boolean isAtStart() {
		if(_timeSpecification.getTimingType() == TimingType.DATA_INDEX) {
			long dataIndex = _delegate.peekDataIndex();
			return dataIndex == _timeSpecification.getIntervalStart();
		}
		if(_timeSpecification.getTimingType() == TimingType.DATA_TIME) {
			long dataTime = _delegate.peekDataTime();
			return dataTime == _timeSpecification.getIntervalStart();
		}
		if(_timeSpecification.getTimingType() == TimingType.ARCHIVE_TIME) {
			long archiveTime = _delegate.peekArchiveTime();
			return archiveTime == _timeSpecification.getIntervalStart();
		}
		throw new AssertionError("Unbekannter TimingType: " + _timeSpecification.getTimingType());
	}

	@Override
	public void peek(final ContainerDataResult result) throws PersistenceException {
		if(_firstData != null) {
			_firstData.copyTo(result);
			return;
		}
		_delegate.peek(result);
	}

	@Override
	public long peekDataIndex() {
		if(_firstData != null) {
			return _firstData.getDataIndex();
		}
		return _delegate.peekDataIndex();
	}

	@Override
	public long peekDataTime() {
		if(_firstData != null) {
			return _firstData.getDataTime();
		}
		return _delegate.peekDataTime();
	}

	@Override
	public long peekArchiveTime() {
		if(_firstData != null) {
			return _firstData.getArchiveTime();
		}
		return _delegate.peekArchiveTime();
	}

	@Override
	public void remove() throws PersistenceException {
		if(_firstData != null) {
			_firstData = null;
			return;
		}
		_delegate.remove();
		removeIfInvalid();
	}

	@Override
	public boolean isEmpty() {
		return _delegate.isEmpty() && _firstData == null;
	}

	@Override
	public void close() throws PersistenceException {
		_delegate.close();
	}

	@Override
	public long getCurContFileIdx() {
		if(isEmpty()) throw new NoSuchElementException();
		if(_delegate != null) {
			return _delegate.getCurContFileIdx();
		}
		return _firstIdx;
	}
}
